# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Component (Face, Aperture) Classes"""

from __future__ import annotations

from typing import ClassVar, Collection, List, Optional, Set, Tuple, Union

from PHX.model import constructions, geometry
from PHX.model.enums.building import (
    ComponentColor,
    ComponentExposureExterior,
    ComponentFaceOpacity,
    ComponentFaceType,
    ThermalBridgeType,
)


class PhxComponentBase:
    """Base class with id_num counter for Opaque and Aperture Components"""

    _count: ClassVar[int] = 0

    def __init__(self):
        PhxComponentBase._count += 1
        self._id_num: int = PhxComponentBase._count

    @property
    def id_num(self) -> int:
        return self._id_num

    def __str__(self):
        return f"{self.__class__.__name__}(id_num={self.id_num})"


class PhxComponentOpaque(PhxComponentBase):
    """Opaque surface components (wall, roof, floor)."""

    def __init__(self):
        super().__init__()

        self.display_name: str = ""
        self.face_type: ComponentFaceType = ComponentFaceType.WALL
        self.face_opacity: ComponentFaceOpacity = ComponentFaceOpacity.OPAQUE
        self.color_interior: ComponentColor = ComponentColor.EXT_WALL_INNER
        self.color_exterior: ComponentColor = ComponentColor.EXT_WALL_INNER
        self.exposure_exterior: ComponentExposureExterior = ComponentExposureExterior.EXTERIOR
        self.exposure_interior: int = 1
        self.interior_attachment_id: int = -1

        self.assembly: constructions.PhxConstructionOpaque = constructions.PhxConstructionOpaque()
        self.assembly_type_id_num: int = -1

        self.apertures: List[PhxComponentAperture] = []
        self.polygons: List[geometry.PhxPolygon] = []

    def __eq__(self, other: PhxComponentOpaque) -> bool:
        if (
            self.display_name != other.display_name
            or self.face_type != other.face_type
            or self.face_opacity != other.face_opacity
            or self.color_interior != other.color_interior
            or self.color_exterior != other.color_exterior
            or self.exposure_exterior != other.exposure_exterior
            or self.exposure_interior != other.exposure_interior
            or self.interior_attachment_id != other.interior_attachment_id
            or self.assembly_type_id_num != other.assembly_type_id_num
        ):
            return False

        # -- check the apertures
        if len(self.apertures) != len(other.apertures):
            return False
        for this_ap in self.apertures:
            if not any((this_ap == other_ap for other_ap in other.apertures)):
                return False

        # -- check the polygons
        if len(self.polygons) != len(other.polygons):
            return False
        for this_poly in self.polygons:
            if not any((this_poly == other_poly for other_poly in other.polygons)):
                return False

        return True

    @property
    def u_value(self) -> float:
        return self.assembly.u_value

    @property
    def polygon_ids(self) -> Set[int]:
        """Return a Set of all the Polygon-id numbers found in the Component's Polygon group."""
        return {polygon.id_num for polygon in self.polygons}

    @property
    def unique_key(self) -> str:
        """Returns a unique text key,. Useful for sorting / grouping / merging components."""
        return (
            f"{self.face_type.value}-{self.face_opacity.value}-{self.exposure_interior}-{self.interior_attachment_id}-"
            f"{self.exposure_exterior.value}-{self.assembly_type_id_num}"
        )

    @property
    def is_shade(self) -> bool:
        if self.face_opacity != ComponentFaceOpacity.OPAQUE:
            return False
        if self.exposure_interior != -1:
            return False
        return True

    @property
    def is_above_grade_wall(self) -> bool:
        if self.face_type != ComponentFaceType.WALL:
            return False
        if self.exposure_exterior != ComponentExposureExterior.EXTERIOR:
            return False
        return True

    @property
    def is_below_grade_wall(self) -> bool:
        if self.face_type != ComponentFaceType.WALL:
            return False
        if self.exposure_exterior != ComponentExposureExterior.GROUND:
            return False
        return True

    @property
    def is_above_grade_floor(self) -> bool:
        if self.face_type != ComponentFaceType.FLOOR:
            return False
        if self.exposure_exterior != ComponentExposureExterior.EXTERIOR:
            return False
        return True

    @property
    def is_below_grade_floor(self) -> bool:
        if self.face_type != ComponentFaceType.FLOOR:
            return False
        if self.exposure_exterior != ComponentExposureExterior.GROUND:
            return False
        return True

    @property
    def is_roof(self) -> bool:
        if self.face_type != ComponentFaceType.ROOF_CEILING:
            return False
        return True

    def add_polygons(self, _input: Union[Collection[geometry.PhxPolygon], geometry.PhxPolygon]) -> None:
        """Adds a new Polygon or Polygons to the Component's collection.

        Arguments:
        ----------
            * _input (Union[Collection[geometry.PhxPolygon], geometry.PhxPolygon]): The polygon or
                polygons to add to the component's collection.

        Returns:
        --------
            * None
        """
        if not isinstance(_input, Collection):
            _input = (_input,)

        for polygon in _input:
            self.polygons.append(polygon)

    @property
    def aperture_ids(self) -> Set[int]:
        """Return a Set of all the Aperture-id numbers found in the Component's Aperture group."""
        return {aperture.id_num for aperture in self.apertures}

    @property
    def aperture_elements(self) -> List[PhxApertureElement]:
        """Return a list of all the Aperture Elements found in the Component's Aperture group."""
        return [element for aperture in self.apertures for element in aperture.elements]

    def __add__(self, other: PhxComponentOpaque) -> PhxComponentOpaque:
        """Merge with another Component into a single new Component.

        Arguments:
        ----------
            * other (PhxComponentOpaque): The other PhxComponentOpaque to merge with.

        Returns:
        --------
            * (PhxComponentOpaque): A new Component with attributes merged.
        """
        new_compo = self.__class__()
        for attr_name, attr_val in vars(self).items():
            if attr_name.startswith("_"):
                continue
            setattr(new_compo, attr_name, attr_val)

        new_compo.display_name = f"{self.face_type.name} [{self.assembly.display_name}]"
        new_compo.polygons = self.polygons + other.polygons
        for phx_aperture in new_compo.apertures:
            phx_aperture.host = new_compo
        for phx_aperture in other.apertures:
            new_compo.add_aperture(phx_aperture)

        return new_compo

    def add_aperture(self, _aperture: PhxComponentAperture) -> None:
        """Add a new child PhxComponentAperture to the Component.

        Arguments:
        ----------
            * _aperture: (PhxComponentAperture): The new PhxComponentAperture to
                add as a child.
        Returns:
        --------
            * None
        """
        if _aperture.id_num not in self.aperture_ids:
            _aperture.host = self
            self.apertures.append(_aperture)

    def get_host_polygon_by_child_id_num(self, _id_num: int) -> geometry.PhxPolygon:
        """Return a single Polygon from the collection if it has the specified ID as a 'child'.

        If the specified ID number is not found, an Exception is raised.

        Arguments:
        ----------
            * _id_num: (int) The Polygon id-number to search the Component's collection for.
        Returns:
        -------
            * (PhxPolygon): The PhxPolygon with the specified id-number.
        """
        for polygon in self.polygons:
            if _id_num in polygon.child_polygon_ids:
                return polygon
        raise Exception(f"Error: Cannot find a host polygon for the child id_num: {_id_num}")

    def get_aperture_polygon_by_id_num(self, _id_num: int) -> geometry.PhxPolygon:
        """Return a single Polygon from the collection if it has the specified ID as a 'child'.

        If the specified ID number is not found, an Exception is raised.

        Arguments:
        ----------
            * _id_num: (int) The Polygon id-number to search the Component's collection for.
        Returns:
        -------
            * (PhxPolygon): The PhxPolygon with the specified id-number.
        """
        for aperture in self.apertures:
            for polygon in aperture.polygons:
                if polygon.id_num == _id_num:
                    return polygon
        raise Exception(f"Error: Cannot find an aperture polygon for the id_num: {_id_num}")

    def get_aperture_element_by_polygon_id_num(self, _id_num: int) -> PhxApertureElement:
        """Return a single Aperture Element from the collection if it has the specified ID.

        If the specified ID number is not found, an Exception is raised.

        Arguments:
        ----------
            * _id_num: (int) The Polygon id-number to search the Component's collection for.
        Returns:
        -------
            * (PhxApertureElement): The PhxApertureElement with the specified id-number.
        """
        for aperture in self.apertures:
            for element in aperture.elements:
                if not element.polygon:
                    continue

                if element.polygon.id_num == _id_num:
                    return element
        raise Exception(f"Error: Cannot find an aperture element for the id_num: {_id_num}")

    def set_assembly_type(self, _phx_construction: constructions.PhxConstructionOpaque):
        """Set the Assembly Type for the Component.

        Arguments:
        ----------
            * _phx_construction (constructions.PhxConstructionOpaque): The Construction to
                set as the Assembly Type.
        Returns:
        --------
            * None
        """
        self.assembly = _phx_construction
        self.assembly_type_id_num = _phx_construction.id_num

    def get_total_gross_component_area(self) -> float:
        """Return the total Gross wall area of the Component (ignoring any nested apertures)."""
        return sum([polygon.area for polygon in self.polygons])

    def get_total_aperture_area(self) -> float:
        """Return the total Gross aperture area of the Component."""
        return sum([aperture.get_total_aperture_area() for aperture in self.apertures])

    def get_total_net_component_area(self) -> float:
        """Return the total net area of the Component (gross - apertures)."""
        return self.get_total_gross_component_area() - self.get_total_aperture_area()


class PhxApertureShadingDimensions(PhxComponentBase):
    """PHPP old-style shading dimensions data.

    ### Horizon Objects:
    - d_hori = Distance away from glass
    - h_hori = Height of the shading from glass bottom edge

    ### Shading to the sides (reveal):
    - d_reveal = Reveal distance 'over' from glass
    - o_reveal = Reveal depth from glass

    ### Overhand Shading:
    - d_over = Overhang height 'up' from glass top edge
    - o_over = Overhang depth from glass
    """

    def __init__(self) -> None:
        super().__init__()

        # -- Horizon Objects
        self.d_hori: Optional[float] = None  # Vertical distance to the shading top
        self.h_hori: Optional[float] = None  # Horizontal distance to the shading

        # -- Shading to the sides (reveal)
        self.d_reveal: Optional[float] = None  # Reveal distance 'over' from glass
        self.o_reveal: Optional[float] = None  # Reveal depth from glass

        # -- Overhand Shading
        self.d_over: Optional[float] = None  # Overhang height 'up' from glass
        self.o_over: Optional[float] = None  # Overhang Depth from glass


class PhxApertureElement(PhxComponentBase):
    """A single sash / element of an Aperture Component."""

    def __init__(self, _host: PhxComponentAperture):
        super().__init__()

        self.host: PhxComponentAperture = _host
        self.display_name: str = ""
        self.polygon: Union[geometry.PhxPolygonRectangular, geometry.PhxPolygon, None] = None
        self.winter_shading_factor: float = 0.75
        self.summer_shading_factor: float = 0.75
        self.shading_dimensions = PhxApertureShadingDimensions()

    @property
    def area(self) -> float:
        """Return the area of the element's polygon."""
        if self.polygon is None:
            return 0.0
        return self.polygon.area

    @property
    def width(self) -> float:
        if not self.polygon:
            return 0.0

        try:
            return self.polygon.width  # type: ignore
        except AttributeError:
            return self.polygon.perimeter_length() / 4

    @property
    def height(self) -> float:
        if not self.polygon:
            return 0.0

        try:
            return self.polygon.height  # type: ignore
        except AttributeError:
            return self.polygon.perimeter_length() / 4

    @property
    def frame_area(self) -> float:
        """Return the area of the frame in the Aperture Element."""
        if not self.polygon:
            return 1.0

        # -- Get the frame widths
        window_type = self.host.window_type
        frame_w_top = window_type.frame_top.width
        frame_w_bottom = window_type.frame_bottom.width
        frame_w_left = window_type.frame_left.width
        frame_w_right = window_type.frame_right.width

        # -- Get the frame lengths (remove the top/bottom frame width from the sides)
        frame_l_top = self.width
        frame_l_bottom = self.width
        frame_l_left = self.height - self.height / self.height * (frame_w_top + frame_w_bottom)
        frame_l_right = self.height - self.height / self.height * (frame_w_top + frame_w_bottom)
        # -- Calc. the frame area
        frame_area = (
            +(frame_l_top * frame_w_top)
            + (frame_l_bottom * frame_w_bottom)
            + (frame_l_left * frame_w_left)
            + (frame_l_right * frame_w_right)
        )

        return frame_area

    @property
    def frame_factor(self) -> float:
        """Return the % of the Aperture Element which is frame (as opposed to glazing)."""
        try:
            return self.frame_area / self.area
        except ZeroDivisionError:
            return 0.0

    @property
    def glazing_area(self) -> float:
        """Return the area of the glazing in the Aperture Element."""
        return self.area * self.glazing_factor

    @property
    def glazing_factor(self) -> float:
        """Return the % of the Aperture Element which is glazing (as opposed to frame)."""
        return 1 - self.frame_factor

    def is_equivalent(self, other: PhxApertureElement) -> bool:
        """Return True if the two elements are equivalent."""
        TOLERANCE = 0.001
        if (
            abs(self.winter_shading_factor - other.winter_shading_factor) > TOLERANCE
            or abs(self.summer_shading_factor - other.summer_shading_factor) > TOLERANCE
        ):
            return False

        if self.polygon != other.polygon:
            return False

        return True

    def scale(self, _scale_factor: float) -> None:
        """Scale the element's polygon by the specified factor."""
        if self.polygon:
            self.polygon.scale(_scale_factor)


class PhxComponentAperture(PhxComponentBase):
    """An Aperture (window, door) component with one or more 'element' (sash)."""

    def __init__(self, _host: PhxComponentOpaque) -> None:
        super().__init__()

        self.host = _host

        self.display_name: str = ""
        self.face_type: ComponentFaceType = ComponentFaceType.WINDOW
        self.face_opacity: ComponentFaceOpacity = ComponentFaceOpacity.TRANSPARENT
        self.color_interior: ComponentColor = ComponentColor.WINDOW
        self.color_exterior: ComponentColor = ComponentColor.WINDOW
        self.exposure_interior: int = 1
        self.exposure_exterior: ComponentExposureExterior = ComponentExposureExterior.EXTERIOR
        self.interior_attachment_id: int = -1

        self.window_type: constructions.PhxConstructionWindow = constructions.PhxConstructionWindow()

        self.variant_type_name: str = "_unnamed_type_"
        self._install_depth: float = 0.1016  # m
        self._default_monthly_shading_correction_factor: float = 0.5

        self.elements: List[PhxApertureElement] = []

    @property
    def install_depth(self) -> float:
        """Return the installation depth of the Aperture."""
        return self._install_depth

    @install_depth.setter
    def install_depth(self, _depth: Optional[float]) -> None:
        """Set the installation depth of the Aperture."""
        if _depth != None:
            self._install_depth = _depth

    @property
    def average_shading_d_reveal(self) -> Optional[float]:
        """The average shading 'reveal' distance of the Aperture's Elements from the glass."""
        if not all(e.shading_dimensions.d_reveal for e in self.elements):
            return None

        try:
            total_d_reveal = sum(e.shading_dimensions.d_reveal or 0.0 for e in self.elements)
            return total_d_reveal / len(self.elements)
        except ZeroDivisionError:
            return 0.0

    @property
    def default_monthly_shading_correction_factor(self) -> float:
        """Return the default monthly shading correction factor."""
        return self._default_monthly_shading_correction_factor

    @default_monthly_shading_correction_factor.setter
    def default_monthly_shading_correction_factor(self, _factor: Optional[float]) -> None:
        """Set the default monthly shading correction factor."""
        if _factor != None:
            self._default_monthly_shading_correction_factor = _factor

    @property
    def shade_type_id_num(self) -> int:
        """Return the ID-Number of the Component Construction's Shade-Type, or -1 if None."""
        return self.window_type._id_num_shade

    @property
    def window_type_id_num(self) -> int:
        return self.window_type.id_num

    @property
    def polygons(
        self,
    ) -> List[Union[geometry.PhxPolygonRectangular, geometry.PhxPolygon]]:
        return [e.polygon for e in self.elements if e.polygon]

    @property
    def polygon_ids(self) -> Set[int]:
        """Return a Set of all the Polygon-id numbers found in the Component's Polygon group."""
        return {polygon.id_num for polygon in self.polygons}

    @property
    def polygon_ids_sorted(self) -> Tuple[int, ...]:
        """Return a Tuple of all the Polygon-id numbers found in the Component's Polygon group, sorted."""
        return tuple(sorted(self.polygon_ids))

    @property
    def unique_key(self) -> str:
        """Returns a unique text key,. Useful for sorting / grouping / merging components."""
        attributes = {
            "face_type": self.face_type.value,
            "face_opacity": self.face_opacity.value,
            "exposure_interior": self.exposure_interior,
            "interior_attachment_id": self.interior_attachment_id,
            "exposure_exterior": self.exposure_exterior.value,
            "window_type_id_num": self.window_type_id_num,
            "shade_type_id_num": self.shade_type_id_num,
            "variant_type_name": self.variant_type_name,
            "default_monthly_shading_correction_factor": self.default_monthly_shading_correction_factor,
        }
        return "-".join(f"{v}" for k, v in attributes.items())

    def add_elements(self, _elements: Collection[PhxApertureElement]) -> None:
        """Add one or more new 'Elements' (Sashes) to the Aperture"""
        for element in _elements:
            self.elements.append(element)

    def add_element(self, _element: PhxApertureElement) -> None:
        """Add a new 'Element' (Sash) to the Aperture"""
        self.elements.append(_element)

    def set_window_type(self, _window_type: constructions.PhxConstructionWindow) -> None:
        """Set the Component's Window Type."""
        self.window_type = _window_type

    def __add__(self, other: PhxComponentAperture) -> PhxComponentAperture:
        """Merge with another Component into a single new Component.

        Arguments:
        ----------
            * other (PhxComponentAperture): The other PhxComponentAperture to merge with.

        Returns:
        --------
            * (PhxComponentAperture): A new Component with attributes merged.
        """
        new_compo = self.__class__(_host=self.host)

        # -- Copy the basic attribute values over
        for attr_name, attr_val in vars(self).items():
            if attr_name.startswith("_"):
                continue  # Ignore private attributes
            setattr(new_compo, attr_name, attr_val)

        # -- Get the Properties as well
        new_compo.install_depth = self.install_depth
        new_compo.default_monthly_shading_correction_factor = self.default_monthly_shading_correction_factor

        if self.window_type.display_name == other.window_type.display_name:
            new_compo.display_name = self.window_type.display_name
        else:
            new_compo.display_name = "Merged_Aperture_Component"

        new_compo.elements = self.elements + other.elements
        for element in new_compo.elements:
            element.host = new_compo

        return new_compo

    def __eq__(self, other: PhxComponentAperture) -> bool:
        if (
            self.display_name != other.display_name
            or self.face_type != other.face_type
            or self.face_opacity != other.face_opacity
            or self.color_interior != other.color_interior
            or self.color_exterior != other.color_exterior
            or self.exposure_exterior != other.exposure_exterior
            or self.exposure_interior != other.exposure_interior
            or self.interior_attachment_id != other.interior_attachment_id
            or self.window_type_id_num != other.window_type_id_num
            or self.variant_type_name != other.variant_type_name
            or abs(self.install_depth - other.install_depth) > 0.001
            or abs(self.default_monthly_shading_correction_factor - other.default_monthly_shading_correction_factor)
            > 0.001
        ):
            return False

        # -- check the elements
        if len(self.elements) != len(other.elements):
            return False
        for this_el in self.elements:
            if not any((this_el.is_equivalent(other_el) for other_el in other.elements)):
                return False

        # -- check the polygons
        if len(self.polygons) != len(other.polygons):
            return False
        for this_poly in self.polygons:
            if not any((this_poly == other_poly for other_poly in other.polygons)):
                return False

        return True

    def get_total_aperture_area(self) -> float:
        """Return the total Window Area of the Component."""
        return sum([polygon.area for polygon in self.polygons])

    def scale(self, _scale_factor: float = 1.0) -> None:
        """Scale the Component's size by the given factor."""
        for element in self.elements:
            element.scale(_scale_factor)


class PhxComponentThermalBridge(PhxComponentBase):
    """A single Thermal Bridge Element."""

    def __init__(self):
        super().__init__()

        self._identifier: Optional[str] = ""
        self._quantity: Optional[float] = 0.0
        self._group_type: Optional[ThermalBridgeType] = ThermalBridgeType.AMBIENT
        self._display_name: Optional[str] = ""
        self._psi_value: Optional[float] = 0.1
        self._fRsi_value: Optional[float] = 0.75
        self._length: Optional[float] = 0.0

    @property
    def identifier(self) -> Optional[str]:
        return self._identifier

    @identifier.setter
    def identifier(self, value: Optional[str]) -> None:
        if value is not None:
            self._identifier = value

    @property
    def quantity(self) -> Optional[float]:
        return self._quantity

    @quantity.setter
    def quantity(self, value: Optional[float]) -> None:
        if value is not None:
            self._quantity = value

    @property
    def group_type(self) -> Optional[ThermalBridgeType]:
        return self._group_type

    @group_type.setter
    def group_type(self, value: Optional[ThermalBridgeType]) -> None:
        if value is not None:
            self._group_type = value

    @property
    def group_number(self) -> int:
        if self.group_type:
            return self.group_type.value
        else:
            return 15

    @property
    def display_name(self) -> Optional[str]:
        return self._display_name

    @display_name.setter
    def display_name(self, value: Optional[str]) -> None:
        if value is not None:
            self._display_name = value

    @property
    def psi_value(self) -> Optional[float]:
        return self._psi_value

    @psi_value.setter
    def psi_value(self, value: Optional[float]) -> None:
        if value is not None:
            self._psi_value = value

    @property
    def fRsi_value(self) -> Optional[float]:
        return self._fRsi_value

    @fRsi_value.setter
    def fRsi_value(self, value: Optional[float]) -> None:
        if value is not None:
            self._fRsi_value = value

    @property
    def length(self) -> Optional[float]:
        return self._length

    @length.setter
    def length(self, value: Optional[float]) -> None:
        if value is not None:
            self._length = value

    @property
    def unique_key(self) -> str:
        """Returns a unique text key,. Useful for sorting / grouping / merging components."""
        return f"{self.group_number}-{self.psi_value :.4f}-{self.display_name}"

    def __add__(self, other: PhxComponentThermalBridge) -> PhxComponentThermalBridge:
        """Merge with another Component into a single new Component.

        Arguments:
        ----------
            * other (PhxComponentThermalBridge): The other PhxComponentThermalBridge to merge with.

        Returns:
        --------
            * (PhxComponentThermalBridge): A new Component with attributes merged.
        """
        new_compo = self.__class__()

        # -- Copy the basic attribute values over
        new_compo.identifier = self.identifier
        new_compo.display_name = self.display_name
        new_compo.group_type = self.group_type
        new_compo.fRsi_value = self.fRsi_value
        new_compo.quantity = 1

        # -- Calculate the new length
        _self_total_len = (self.length or 0) * (self.quantity or 0)
        _other_total_len = (other.length or 0) * (other.quantity or 0)
        new_compo.length = _self_total_len + _other_total_len

        # -- Calculate the new psi value
        _self_weighted_psi_value = _self_total_len * (self.psi_value or 0.0)
        _other_weighted_psi_value = _other_total_len * (other.psi_value or 0.0)
        _total_weighted_psi_value = _self_weighted_psi_value + _other_weighted_psi_value
        try:
            new_compo.psi_value = _total_weighted_psi_value / new_compo.length
        except ZeroDivisionError:
            new_compo.psi_value = 0.0

        return new_compo

    def __eq__(self, other: PhxComponentThermalBridge) -> bool:
        TOLERANCE = 0.001
        if (
            self.group_type != other.group_type
            or self.display_name != other.display_name
            or abs((self.psi_value or 0) - (other.psi_value or 0)) > TOLERANCE
            or abs((self.fRsi_value or 0) - (other.fRsi_value or 0)) > TOLERANCE
            or abs((self.length or 0) - (other.length or 0)) > TOLERANCE
        ):
            return False
        return True



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Space (Room) Class"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import ClassVar

from PHX.model.programs.lighting import PhxProgramLighting
from PHX.model.programs.occupancy import PhxProgramOccupancy
from PHX.model.programs.ventilation import PhxProgramVentilation


def area_weighted_clear_height(space_a: PhxSpace, space_b: PhxSpace) -> float:
    """Returns the area-weighted clear-height of two spaces."""
    try:
        weighted_height_a = space_a.clear_height * space_a.floor_area
        weighted_height_b = space_b.clear_height * space_b.floor_area
        total_floor_area = space_a.floor_area + space_b.floor_area
        return (weighted_height_a + weighted_height_b) / total_floor_area
    except ZeroDivisionError:
        return 0.0


def spaces_are_not_addable(space_a: PhxSpace, space_v: PhxSpace) -> bool:
    """Returns True if space_a can safely be added to space_b."""
    return any(
        (
            space_a.wufi_type != space_v.wufi_type,
            space_a.vent_unit_id_num != space_v.vent_unit_id_num,
        )
    )


@dataclass
class PhxSpace:
    _count: ClassVar[int] = 0

    id_num: int = field(init=False, default=0)
    display_name: str = "Unnamed_Space"
    wufi_type: int = 99  # User Determined
    quantity: int = 1
    floor_area: float = 0.0
    weighted_floor_area: float = 0.0
    net_volume: float = 0.0
    clear_height: float = 2.5

    # -- Ventilation Unit (ERV) number
    vent_unit_id_num: int = 0
    vent_unit_display_name: str = ""

    # -- Programs
    ventilation: PhxProgramVentilation = field(default_factory=PhxProgramVentilation)
    occupancy: PhxProgramOccupancy = field(default_factory=PhxProgramOccupancy)
    lighting: PhxProgramLighting = field(default_factory=PhxProgramLighting)

    electric_equipment = None

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def peak_occupancy(self) -> float:
        """Returns the peak occupancy for the space (Num. of people)."""
        return self.occupancy.load.people_per_m2 * self.floor_area

    @peak_occupancy.setter
    def peak_occupancy(self, value: float) -> None:
        """Sets the peak occupancy for the space (Num. of people)."""
        try:
            self.occupancy.load.people_per_m2 = value / self.floor_area
        except ZeroDivisionError:
            self.occupancy.load.people_per_m2 = 0.0

    @property
    def has_ventilation_airflow(self) -> bool:
        """Returns True if the space has ventilation airflow."""
        return self.ventilation.has_ventilation_airflow

    def __add__(self, other: PhxSpace) -> PhxSpace:
        """Adds two spaces together.

        This is used to report out 'grouped' spaces for Phius. In almost all
        other cases, spaces should be kept separated. Note that the occupancy,
        lighting, and ventilation schedules are NOT merged, however the
        ventilation loads ARE added together.
        """
        if spaces_are_not_addable(self, other):
            raise ValueError(f"Space {self.display_name} cannot be added to space {other.display_name}.")

        new_space = PhxSpace(
            display_name=self.vent_unit_display_name,
            wufi_type=self.wufi_type,
            quantity=1,
            floor_area=self.floor_area + other.floor_area,
            weighted_floor_area=self.weighted_floor_area + other.weighted_floor_area,
            net_volume=self.net_volume + other.net_volume,
            clear_height=area_weighted_clear_height(self, other),
            vent_unit_id_num=self.vent_unit_id_num,
            vent_unit_display_name=self.vent_unit_display_name,
            ventilation=self.ventilation,
            occupancy=self.occupancy,
            lighting=self.lighting,
        )
        new_space.ventilation.load = self.ventilation.load + other.ventilation.load
        return new_space



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Mechanical Cooling Devices and Parameters."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, Union

from PHX.model.hvac import _base

# -- Cooling Parameter Types --------------------------------------------------


@dataclass
class PhxCoolingVentilationParams(_base.PhxMechanicalDeviceParams):
    used: bool = False
    single_speed: bool = False
    min_coil_temp: float = 12  # C
    capacity: float = 10  # kW
    annual_COP: float = 4  # W/W

    @property
    def total_system_perf_ratio(self):
        return 1 / self.annual_COP

    def __add__(self, other: PhxCoolingVentilationParams) -> PhxCoolingVentilationParams:
        if not other:
            return self

        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))
        new_obj.used = any([self.used, other.used])
        new_obj.single_speed = any([self.single_speed, other.single_speed])
        new_obj.min_coil_temp = (self.min_coil_temp + other.min_coil_temp) / 2
        new_obj.capacity = (self.capacity + other.capacity) / 2
        new_obj.annual_COP = (self.annual_COP + other.annual_COP) / 2
        return new_obj


@dataclass
class PhxCoolingRecirculationParams(_base.PhxMechanicalDeviceParams):
    used: bool = False
    single_speed: bool = False
    min_coil_temp: float = 12  # C
    capacity: float = 10  # kW
    annual_COP: float = 4  # W/W
    flow_rate_m3_hr: float = 100
    flow_rate_variable: bool = True

    @property
    def total_system_perf_ratio(self):
        return 1 / self.annual_COP

    def __add__(self, other: PhxCoolingRecirculationParams) -> PhxCoolingRecirculationParams:
        if not other:
            return self

        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))
        new_obj.used = any([self.used, other.used])
        new_obj.single_speed = any([self.single_speed, other.single_speed])
        new_obj.min_coil_temp = (self.min_coil_temp + other.min_coil_temp) / 2
        new_obj.flow_rate_m3_hr = (self.flow_rate_m3_hr + other.flow_rate_m3_hr) / 2
        new_obj.capacity = (self.capacity + other.capacity) / 2
        new_obj.flow_rate_variable = any([self.flow_rate_variable, other.flow_rate_variable])
        new_obj.annual_COP = (self.annual_COP + other.annual_COP) / 2
        return new_obj


@dataclass
class PhxCoolingDehumidificationParams(_base.PhxMechanicalDeviceParams):
    used: bool = False
    annual_COP: float = 4  # W/W
    useful_heat_loss: bool = False

    @property
    def total_system_perf_ratio(self):
        return 1 / self.annual_COP

    def __add__(self, other: PhxCoolingDehumidificationParams) -> PhxCoolingDehumidificationParams:
        if not other:
            return self

        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))
        new_obj.used = any([self.used, other.used])
        new_obj.useful_heat_loss = any([self.useful_heat_loss, other.useful_heat_loss])
        new_obj.annual_COP = (self.annual_COP + other.annual_COP) / 2
        return new_obj


@dataclass
class PhxCoolingPanelParams(_base.PhxMechanicalDeviceParams):
    used: bool = False
    annual_COP: float = 4  # W/W

    @property
    def total_system_perf_ratio(self):
        return 1 / self.annual_COP

    def __add__(self, other: PhxCoolingPanelParams) -> PhxCoolingPanelParams:
        if not other:
            return self

        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))
        new_obj.used = any([self.used, other.used])
        new_obj.annual_COP = (self.annual_COP + other.annual_COP) / 2
        return new_obj


AnyPhxCoolingParamsType = Union[
    PhxCoolingVentilationParams,
    PhxCoolingRecirculationParams,
    PhxCoolingDehumidificationParams,
    PhxCoolingPanelParams,
]

# -- Param Collection ---------------------------------------------------------


@dataclass
class PhxCoolingParams:
    ventilation: PhxCoolingVentilationParams = field(default_factory=PhxCoolingVentilationParams)
    recirculation: PhxCoolingRecirculationParams = field(default_factory=PhxCoolingRecirculationParams)
    dehumidification: PhxCoolingDehumidificationParams = field(default_factory=PhxCoolingDehumidificationParams)
    panel: PhxCoolingPanelParams = field(default_factory=PhxCoolingPanelParams)

    def __bool__(self) -> bool:
        return any([self.ventilation, self.recirculation, self.dehumidification, self.panel])



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Pydantic Model for reading in WUFI-XML file format."""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from honeybee_ph_standards.sourcefactors import factors, phius_CO2_factors, phius_source_energy_factors
from ph_units.converter import convert
from pydantic import BaseModel, validator

from PHX.from_WUFI_XML import wufi_file_types as wufi_unit
from PHX.from_WUFI_XML.read_WUFI_XML_file import Tag
from PHX.model.enums.phius_certification import PhiusCertificationProgram

# ------------------------------------------------------------------------------
# -- Unit Types

TagList = List[Tag]
TagDict = Dict[str, Tag]
UnitDict = Dict[str, Any]


def unpack_xml_tag(_input: Union[TagList, TagDict, UnitDict, Tag]) -> Union[TagList, TagDict, UnitDict, str, None]:
    """This validator should run first before any unit-conversion. This will
    unpack the Tag object and return the .text, or a dict with the unit and the
    value if it is a unit-type.


    _value can be:
        * List: [...]
        * Dict: {'Year': Tag(...), 'Month': Tag(...), 'Day': Tag(...), 'Hour': Tag(...), 'Minutes': Tag(...)}
        * Tag (no unit):  Tag(text='New York, NY', tag='Customer_Locality', attrib={})
        * Tag (with unit):  Tag(text='7.5', tag='BeginUtilization', attrib={'unit': 'hr'})
    """

    if isinstance(_input, Tag):
        # print(f"  <{type(_input).__name__}> {_input}")
        # -- If it is a unit-type, pass back a dict with the unit and the value
        # -- this will get unpacked and used by the unit-type converter.
        unit_type = getattr(_input, "attrib", {}).get("unit", None)
        if unit_type:
            return {"value": _input.text, "unit_type": unit_type}

        # -- Otherwise, just pass back the .text attribute of the Tag
        if _input.text == "None":
            return None
        else:
            return _input.text
    else:
        # print(f"{type(_input).__name__}")
        return _input


# ------------------------------------------------------------------------------
# -- Climate


class WufiCO2FactorsUserDef(BaseModel):
    __root__: Any

    @validator("__root__", pre=True)
    def unpack_xml_tag_name(cls, v):
        # -- Have to do it this way cus' the WUFI file structure is a mess
        # Input is a dict like: '{'PEF0': Tag(text='1.1', tag='PEF0', attrib={'unit': 'Btu/Btu'})}'
        input_tag = list(v.values())[-1]
        input_dict: Dict = unpack_xml_tag(input_tag)  # type: ignore
        result = convert(input_dict["value"], str(input_dict["unit_type"]), "G/KWH")
        if not result:
            msg = f"Error converting {input_dict['value']} from {input_dict['unit_type']} to G/KWH"
            raise Exception(msg)
        return float(result)


class WufiPEFactorsUserDef(BaseModel):
    __root__: Any

    @validator("__root__", pre=True)
    def unpack_xml_tag_name(cls, v):
        # -- Have to do it this way cus' the WUFI file structure is a mess
        # Input is a dict like: '{'PEF0': Tag(text='1.1', tag='PEF0', attrib={'unit': 'Btu/Btu'})}'
        input_tag = list(v.values())[-1]
        input_dict: Dict = unpack_xml_tag(input_tag)  # type: ignore
        result = convert(input_dict["value"], str(input_dict["unit_type"]), "KWH/KWH")
        if not result:
            msg = f"Error converting {input_dict['value']} from {input_dict['unit_type']} to KWH/KWH"
            raise Exception(msg)
        return float(result)


class WufiMonthlyClimateTemp_Item(BaseModel):
    Item: Optional[wufi_unit.DegreeC] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiMonthlyClimateRadiation_Item(BaseModel):
    Item: Optional[wufi_unit.kWh_per_M2] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiPH_ClimateLocation(BaseModel):
    Selection: int
    Name: Optional[str] = None
    Comment: Optional[str] = None
    DailyTemperatureSwingSummer: wufi_unit.DegreeDeltaK
    AverageWindSpeed: wufi_unit.M_per_Second

    Latitude: wufi_unit.CardinalDegrees
    Longitude: wufi_unit.CardinalDegrees
    dUTC: int

    HeightNNWeatherStation: wufi_unit.M
    HeightNNBuilding: wufi_unit.M

    ClimateZone: int
    GroundThermalConductivity: wufi_unit.Watts_per_MK
    GroundHeatCapacitiy: wufi_unit.Joule_per_KGK
    GroundDensity: wufi_unit.KG_per_M3
    DepthGroundwater: wufi_unit.M
    FlowRateGroundwater: wufi_unit.M_per_Day

    TemperatureMonthly: List[Optional[WufiMonthlyClimateTemp_Item]]
    DewPointTemperatureMonthly: List[Optional[WufiMonthlyClimateTemp_Item]]
    SkyTemperatureMonthly: List[Optional[WufiMonthlyClimateTemp_Item]]

    NorthSolarRadiationMonthly: List[Optional[WufiMonthlyClimateRadiation_Item]]
    EastSolarRadiationMonthly: List[Optional[WufiMonthlyClimateRadiation_Item]]
    SouthSolarRadiationMonthly: List[Optional[WufiMonthlyClimateRadiation_Item]]
    WestSolarRadiationMonthly: List[Optional[WufiMonthlyClimateRadiation_Item]]
    GlobalSolarRadiationMonthly: List[Optional[WufiMonthlyClimateRadiation_Item]]

    TemperatureHeating1: Optional[wufi_unit.DegreeC] = None
    NorthSolarRadiationHeating1: Optional[wufi_unit.Watts_per_M2] = None
    EastSolarRadiationHeating1: Optional[wufi_unit.Watts_per_M2] = None
    SouthSolarRadiationHeating1: Optional[wufi_unit.Watts_per_M2] = None
    WestSolarRadiationHeating1: Optional[wufi_unit.Watts_per_M2] = None
    GlobalSolarRadiationHeating1: Optional[wufi_unit.Watts_per_M2] = None

    TemperatureHeating2: Optional[wufi_unit.DegreeC] = None
    NorthSolarRadiationHeating2: Optional[wufi_unit.Watts_per_M2] = None
    EastSolarRadiationHeating2: Optional[wufi_unit.Watts_per_M2] = None
    SouthSolarRadiationHeating2: Optional[wufi_unit.Watts_per_M2] = None
    WestSolarRadiationHeating2: Optional[wufi_unit.Watts_per_M2] = None
    GlobalSolarRadiationHeating2: Optional[wufi_unit.Watts_per_M2] = None

    TemperatureCooling: Optional[wufi_unit.DegreeC] = None
    NorthSolarRadiationCooling: Optional[wufi_unit.Watts_per_M2] = None
    EastSolarRadiationCooling: Optional[wufi_unit.Watts_per_M2] = None
    SouthSolarRadiationCooling: Optional[wufi_unit.Watts_per_M2] = None
    WestSolarRadiationCooling: Optional[wufi_unit.Watts_per_M2] = None
    GlobalSolarRadiationCooling: Optional[wufi_unit.Watts_per_M2] = None

    TemperatureCooling2: Optional[wufi_unit.DegreeC] = None
    NorthSolarRadiationCooling2: Optional[wufi_unit.Watts_per_M2] = None
    EastSolarRadiationCooling2: Optional[wufi_unit.Watts_per_M2] = None
    SouthSolarRadiationCooling2: Optional[wufi_unit.Watts_per_M2] = None
    WestSolarRadiationCooling2: Optional[wufi_unit.Watts_per_M2] = None
    GlobalSolarRadiationCooling2: Optional[wufi_unit.Watts_per_M2] = None

    SelectionPECO2Factor: int
    PEFactorsUserDef: Optional[List[WufiPEFactorsUserDef]] = None
    CO2FactorsUserDef: Optional[List[WufiCO2FactorsUserDef]] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)

    def set_standard_pe_factors(self, PH_CertificateCriteriaNum: int) -> None:
        """Set the PE-Factors from the Standards-Library based on the PH_CertificateCriteria."""
        self.PEFactorsUserDef = []

        # -- Load in the Factor values from the Standards library
        factor_mapping = {
            PhiusCertificationProgram.DEFAULT: phius_source_energy_factors.factors_2021,
            PhiusCertificationProgram.PHIUS_2015: phius_source_energy_factors.factors_2018,
            PhiusCertificationProgram.PHIUS_2018: phius_source_energy_factors.factors_2018,
            PhiusCertificationProgram.ITALIAN: phius_source_energy_factors.factors_2021,
            PhiusCertificationProgram.PHIUS_2018_CORE: phius_source_energy_factors.factors_2018,
            PhiusCertificationProgram.PHIUS_2018_ZERO: phius_source_energy_factors.factors_2018,
            PhiusCertificationProgram.PHIUS_2021_CORE: phius_source_energy_factors.factors_2021,
            PhiusCertificationProgram.PHIUS_2021_ZERO: phius_source_energy_factors.factors_2021,
        }
        factor_dict = factor_mapping[PhiusCertificationProgram(PH_CertificateCriteriaNum)]
        factor_objects = factors.build_factors_from_library(factor_dict)

        # -- Mimic the XML structure normal data would come in as
        # -- and build up all the PEFactorsUserDef objects
        for i, obj in enumerate(factor_objects):
            obj.unit
            x = {f"PEF{i}": Tag(text=str(obj.value), tag=f"PEF{0}", attrib={"unit": obj.unit})}
            self.PEFactorsUserDef.append(WufiPEFactorsUserDef.parse_obj(x))

    def set_standard_co2_factors(self, PH_CertificateCriteriaNum: int) -> None:
        """Set the CO2-Factors from the Standards-Library based on the PH_CertificateCriteria."""
        self.CO2FactorsUserDef = []

        # -- Load in the Factor values from the Standards library
        factor_mapping = {
            PhiusCertificationProgram.DEFAULT: phius_CO2_factors.factors_2021,
            PhiusCertificationProgram.PHIUS_2015: phius_CO2_factors.factors_2018,
            PhiusCertificationProgram.PHIUS_2018: phius_CO2_factors.factors_2018,
            PhiusCertificationProgram.ITALIAN: phius_CO2_factors.factors_2021,
            PhiusCertificationProgram.PHIUS_2018_CORE: phius_CO2_factors.factors_2018,
            PhiusCertificationProgram.PHIUS_2018_ZERO: phius_CO2_factors.factors_2018,
            PhiusCertificationProgram.PHIUS_2021_CORE: phius_CO2_factors.factors_2021,
            PhiusCertificationProgram.PHIUS_2021_ZERO: phius_CO2_factors.factors_2021,
        }
        factor_dict = factor_mapping[PhiusCertificationProgram(PH_CertificateCriteriaNum)]
        factor_objects = factors.build_factors_from_library(factor_dict)

        # -- Mimic the XML structure normal data would come in as
        # -- and build up all the PEFactorsUserDef objects
        for i, obj in enumerate(factor_objects):
            x = {f"CO2F{i}": Tag(text=str(obj.value), tag=f"CO2F{0}", attrib={"unit": obj.unit})}
            self.CO2FactorsUserDef.append(WufiCO2FactorsUserDef.parse_obj(x))


class WufiClimateLocation(BaseModel):
    Selection: int

    Latitude_DB: Optional[wufi_unit.CardinalDegrees] = None
    Longitude_DB: Optional[wufi_unit.CardinalDegrees] = None
    HeightNN_DB: Optional[wufi_unit.M] = None
    dUTC_DB: Optional[float] = None

    Albedo: int
    GroundReflShort: wufi_unit._Percentage
    GroundReflLong: wufi_unit._Percentage
    GroundEmission: wufi_unit._Percentage
    CloudIndex: wufi_unit._Percentage
    CO2concenration: wufi_unit.MG_per_M3
    Unit_CO2concentration: wufi_unit.PartsPerMillionByVolume
    PH_ClimateLocation: Optional[WufiPH_ClimateLocation] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Geometry


class WufiVertix(BaseModel):
    # Note: Bizarrely, even in IP units WUFI XML, the Vertices are in Meters.
    IdentNr: int
    X: float
    Y: float
    Z: float

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiIdentNr(BaseModel):
    IdentNr: int
    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiPolygon(BaseModel):
    IdentNr: int
    NormalVectorX: float
    NormalVectorY: float
    NormalVectorZ: float
    IdentNrPoints: List[WufiIdentNr]
    IdentNrPolygonsInside: Optional[List[WufiIdentNr]] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiGraphics_3D(BaseModel):
    Vertices: List[WufiVertix]
    Polygons: List[WufiPolygon]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Building and Components


class WufiRoom(BaseModel):
    Name: str
    Type: int
    IdentNrUtilizationPatternVent: int
    IdentNrVentilationUnit: int
    Quantity: int
    AreaRoom: Optional[wufi_unit.M2] = None
    ClearRoomHeight: Optional[wufi_unit.M] = None
    DesignVolumeFlowRateSupply: wufi_unit.M3_per_Hour
    DesignVolumeFlowRateExhaust: wufi_unit.M3_per_Hour

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiLoadPerson(BaseModel):
    Name: str
    IdentNrUtilizationPattern: int
    ChoiceActivityPersons: int
    NumberOccupants: wufi_unit._Float
    FloorAreaUtilizationZone: wufi_unit.M2

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiLoadsLighting(BaseModel):
    Name: str
    RoomCategory: int
    ChoiceLightTransmissionGlazing: int
    LightingControl: int
    WithinThermalEnvelope: bool
    MotionDetector: bool
    FacadeIncludingWindows: bool
    FractionTreatedFloorArea: wufi_unit._Percentage
    DeviationFromNorth: wufi_unit.AngleDegree
    RoomDepth: wufi_unit.M
    RoomWidth: wufi_unit.M
    RoomHeight: wufi_unit.M
    LintelHeight: wufi_unit.M
    WindowWidth: wufi_unit.M
    InstalledLightingPower: wufi_unit.Watts_per_M2
    LightingFullLoadHours: wufi_unit.Hours_per_Year

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiExhaustVent(BaseModel):
    Name: Optional[str] = None
    Type: int
    ExhaustVolumeFlowRate: Optional[wufi_unit.M3_per_Hour] = None
    RunTimePerYear: Optional[wufi_unit.Hours_per_Year] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiHomeDevice(BaseModel):
    # -- Basic
    Comment: Optional[str] = None
    ReferenceQuantity: Optional[int] = None
    Quantity: Optional[int] = None
    InConditionedSpace: Optional[bool] = None
    ReferenceEnergyDemandNorm: Optional[int] = None
    EnergyDemandNorm: Optional[wufi_unit.kWh] = None
    EnergyDemandNormUse: Optional[wufi_unit.kWh] = None
    CEF_CombinedEnergyFactor: Optional[wufi_unit._Percentage] = None
    Type: int

    # -- Cook-tops
    CookingWith: Optional[int] = None

    # -- Lighting
    FractionHightEfficiency: Optional[wufi_unit._Percentage] = None

    # -- Dishwasher
    CookingWith: Optional[int] = None

    # -- Dryer
    Dryer_Choice: Optional[int] = None
    GasConsumption: Optional[wufi_unit.kWh] = None
    EfficiencyFactorGas: Optional[wufi_unit._Percentage] = None
    FieldUtilizationFactorPreselection: Optional[int] = None
    FieldUtilizationFactor: Optional[wufi_unit._Float] = None

    #  -- Washer
    Connection: Optional[int] = None
    UtilizationFactor: Optional[wufi_unit._Float] = None
    CapacityClothesWasher: Optional[wufi_unit.M3] = None
    MEF_ModifiedEnergyFactor: Optional[wufi_unit._Float] = None

    # -- Dishwasher
    DishwasherCapacityPreselection: Optional[int] = None
    DishwasherCapacityInPlace: Optional[wufi_unit._Float] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiIdentNrPolygons(BaseModel):
    IdentNr: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiComponent(BaseModel):
    IdentNr: int
    Name: Optional[str] = None
    Visual: bool
    Type: int
    IdentNrColorI: int
    IdentNrColorE: int
    InnerAttachment: int
    OuterAttachment: int
    IdentNr_ComponentInnerSurface: Optional[int] = None
    IdentNrAssembly: Optional[int] = None
    IdentNrWindowType: Optional[int] = None
    IdentNrPolygons: List[WufiIdentNrPolygons]

    # Window-Specific Attributes
    DepthWindowReveal: Optional[wufi_unit.M] = None
    IdentNrSolarProtection: Optional[int] = None
    IdentNrOverhang: Optional[int] = None
    DefaultCorrectionShadingMonth: Optional[wufi_unit._Percentage] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiThermalBridge(BaseModel):
    Name: Optional[str] = None
    Type: int
    Length: Optional[wufi_unit.M] = None
    PsiValue: Optional[wufi_unit.Watts_per_MK] = None
    IdentNrOptionalClimate: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiZone(BaseModel):
    Name: str
    KindZone: int
    KindAttachedZone: Optional[int] = None
    TemperatureReductionFactorUserDefined: Optional[wufi_unit._Percentage] = None
    IdentNr: int
    RoomsVentilation: Optional[List[WufiRoom]] = None
    LoadsPersonsPH: Optional[List[WufiLoadPerson]] = None
    LoadsLightingsPH: Optional[List[WufiLoadsLighting]] = None
    GrossVolume_Selection: int
    GrossVolume: Optional[wufi_unit.M3] = None
    NetVolume_Selection: int
    NetVolume: Optional[wufi_unit.M3] = None
    FloorArea_Selection: int
    FloorArea: Optional[wufi_unit.M2] = None
    ClearanceHeight_Selection: int
    ClearanceHeight: Optional[wufi_unit.M] = None
    SpecificHeatCapacity_Selection: int
    SpecificHeatCapacity: wufi_unit.Wh_per_M2K
    IdentNrPH_Building: int
    OccupantQuantityUserDef: wufi_unit._Int
    NumberBedrooms: Optional[wufi_unit._Int] = None
    SummerNaturalVentilationDay: Optional[wufi_unit.ACH] = None
    SummerNaturalVentilationNight: Optional[wufi_unit.ACH] = None

    HomeDevice: Optional[List[WufiHomeDevice]] = None
    ExhaustVents: Optional[List[WufiExhaustVent]] = None
    ThermalBridges: Optional[List[WufiThermalBridge]] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiBuilding(BaseModel):
    Components: List[WufiComponent]
    Zones: List[WufiZone]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- HVAC


class WufiTwig(BaseModel):
    Name: str
    IdentNr: int
    PipingLength: Optional[wufi_unit.M] = None
    PipeMaterial: int
    PipingDiameter: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiBranch(BaseModel):
    Name: Optional[str] = None
    IdentNr: int
    PipingLength: Optional[wufi_unit.M] = None
    PipeMaterial: int
    PipingDiameter: int
    Twigs: Optional[List[WufiTwig]] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiTrunc(BaseModel):
    Name: Optional[str] = None
    IdentNr: int
    PipingLength: Optional[wufi_unit.M] = None
    PipeMaterial: int
    PipingDiameter: int
    CountUnitsOrFloors: int
    DemandRecirculation: bool
    Branches: Optional[List[WufiBranch]] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiDistributionDHW(BaseModel):
    CalculationMethodIndividualPipes: int
    DemandRecirculation: bool
    SelectionhotWaterFixtureEff: int
    NumberOfBathrooms: Optional[int] = None
    AllPipesAreInsulated: bool
    SelectionUnitsOrFloors: int
    PipeMaterialSimplifiedMethod: int
    PipeDiameterSimplifiedMethod: Optional[float] = None
    TemperatureRoom_WR: Optional[wufi_unit.DegreeC] = None
    DesignFlowTemperature_WR: Optional[wufi_unit.DegreeC] = None
    DailyRunningHoursCirculation_WR: Optional[wufi_unit.Hour] = None
    LengthCirculationPipes_WR: Optional[wufi_unit.M] = None
    HeatLossCoefficient_WR: Optional[wufi_unit.Watts_per_MK] = None
    LengthIndividualPipes_WR: Optional[wufi_unit.M] = None
    ExteriorPipeDiameter_WR: Optional[wufi_unit.M] = None
    Truncs: Optional[List[WufiTrunc]] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiDistributionCooling(BaseModel):
    CoolingViaRecirculation: Optional[bool] = None
    RecirculatingAirOnOff: Optional[bool] = None
    MaxRecirculationAirCoolingPower: Optional[wufi_unit.KiloWatt] = None
    MinTempCoolingCoilRecirculatingAir: Optional[wufi_unit.DegreeC] = None
    RecirculationCoolingCOP: Optional[wufi_unit._Percentage] = None
    RecirculationAirVolume: Optional[wufi_unit.M3_per_Hour]
    ControlledRecirculationVolumeFlow: bool

    # -- Ventilation Air
    CoolingViaVentilationAir: Optional[bool] = None
    SupplyAirCoolingOnOff: Optional[bool] = None
    SupplyAirCoolinCOP: Optional[wufi_unit._Percentage] = None
    MaxSupplyAirCoolingPower: Optional[wufi_unit.KiloWatt] = None
    MinTemperatureCoolingCoilSupplyAir: Optional[wufi_unit.DegreeC] = None

    # -- Dehumidification
    Dehumidification: Optional[bool] = None
    DehumdificationCOP: Optional[wufi_unit._Percentage] = None
    UsefullDehumidificationHeatLoss: Optional[bool] = None

    # -- Panel
    PanelCooling: Optional[bool] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiSupportiveDevice(BaseModel):
    Name: Optional[str] = None
    Type: int
    Quantity: int
    InConditionedSpace: Optional[bool] = None
    NormEnergyDemand: Optional[wufi_unit.Watts] = None
    Controlled: Optional[bool] = None
    PeriodOperation: Optional[wufi_unit.KiloHours_per_Year] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiAssignedVentUnit(BaseModel):
    IdentNrVentUnit: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiDuct(BaseModel):
    Name: Optional[str] = None
    IdentNr: int
    DuctDiameter: Optional[wufi_unit.MM] = None
    DuctShapeHeight: Optional[wufi_unit.MM] = None
    DuctShapeWidth: Optional[wufi_unit.MM] = None
    DuctLength: Optional[wufi_unit.M] = None
    InsulationThickness: Optional[wufi_unit.MM] = None
    ThermalConductivity: Optional[wufi_unit.Watts_per_MK] = None
    Quantity: Optional[wufi_unit._Int] = None
    DuctType: int
    DuctShape: int
    IsReflective: bool
    AssignedVentUnits: Optional[List[WufiAssignedVentUnit]] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiPHDistribution(BaseModel):
    # TODO DistributionHeating: DistributionHeating
    DistributionDHW: Optional[WufiDistributionDHW] = None
    DistributionCooling: Optional[WufiDistributionCooling] = None
    DistributionVentilation: Optional[List[WufiDuct]] = None
    UseDefaultValues: bool
    DeviceInConditionedSpace: Optional[bool] = None
    SupportiveDevices: Optional[List[WufiSupportiveDevice]] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiZoneCoverage(BaseModel):
    IdentNrZone: int
    CoverageHeating: float
    CoverageCooling: float
    CoverageVentilation: float
    CoverageHumidification: float
    CoverageDehumidification: float

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiPH_Parameters(BaseModel):
    # -- Heat Pump
    InConditionedSpace: Optional[bool] = None
    AuxiliaryEnergy: Optional[wufi_unit.Watts] = None
    AuxiliaryEnergyDHW: Optional[wufi_unit.Watts] = None
    AnnualCOP: Optional[wufi_unit.kWh_per_kWh] = None
    TotalSystemPerformanceRatioHeatGenerator: Optional[wufi_unit._Percentage] = None
    HPType: Optional[int] = None
    RatedCOP1: Optional[wufi_unit._Percentage] = None
    RatedCOP2: Optional[wufi_unit._Percentage] = None
    AmbientTemperature1: Optional[wufi_unit.DegreeC] = None
    AmbientTemperature2: Optional[wufi_unit.DegreeC] = None

    # -- Ventilation
    Quantity: Optional[int] = None
    HumidityRecoveryEfficiency: Optional[wufi_unit._Percentage] = None
    ElectricEfficiency: Optional[wufi_unit.Wh_per_M3] = None
    DefrostRequired: Optional[bool] = None
    FrostProtection: Optional[bool] = None
    TemperatureBelowDefrostUsed: Optional[wufi_unit.DegreeC] = None
    NoSummerBypass: Optional[bool] = None

    # -- HP Water Heater
    HPWH_EF: Optional[wufi_unit._Percentage] = None

    # -- Water Storage
    QauntityWS: Optional[int] = None
    SolarThermalStorageCapacity: Optional[wufi_unit.Liter] = None
    StorageLossesStandby: Optional[wufi_unit.Watts_per_DegreeK] = None
    TotalSolarThermalStorageLosses: Optional[wufi_unit.Watts_per_DegreeK] = None
    InputOption: Optional[int] = None
    AverageHeatReleaseStorage: Optional[wufi_unit.Watts] = None
    TankRoomTemp: Optional[wufi_unit.DegreeC] = None
    TypicalStorageWaterTemperature: Optional[wufi_unit.DegreeC] = None

    # -- PV System
    SelectionLocation: Optional[int] = None
    SelectionOnSiteUtilization: Optional[int] = None
    SelectionUtilization: Optional[int] = None
    ArraySizePV: Optional[float] = None
    PhotovoltaicRenewableEnergy: Optional[wufi_unit.kWh] = None
    OnsiteUtilization: Optional[wufi_unit._Percentage] = None

    # -- Boilers
    EnergySourceBoilerType: Optional[int] = None
    CondensingBoiler: Optional[bool] = None
    BoilerEfficiency30: Optional[wufi_unit._Percentage] = None
    BoilerEfficiencyNominalOutput: Optional[float] = None
    AverageReturnTemperatureMeasured30Load: Optional[float] = None
    AverageBoilerTemperatureDesign70_55: Optional[float] = None
    AverageBoilerTemperatureDesign55_45: Optional[float] = None
    AverageBoilerTemperatureDesign35_28: Optional[float] = None
    StandbyHeatLossBoiler70: Optional[float] = None
    MaximalBoilerPower: Optional[float] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiDHW_Parameters(BaseModel):
    CoverageWithinSystem: wufi_unit._Percentage
    Unit: float
    Selection: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiHeating_Parameters(BaseModel):
    CoverageWithinSystem: wufi_unit._Percentage
    Unit: float
    Selection: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiCooling_Parameters(BaseModel):
    CoverageWithinSystem: Optional[wufi_unit._Percentage] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiVentilation_Parameters(BaseModel):
    CoverageWithinSystem: Optional[wufi_unit._Percentage] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiDevice(BaseModel):
    Name: Optional[str] = None
    IdentNr: int
    SystemType: int
    TypeDevice: int

    UsedFor_Heating: bool
    UsedFor_DHW: bool
    UsedFor_Cooling: bool
    UsedFor_Ventilation: bool
    UsedFor_Humidification: bool
    UsedFor_Dehumidification: bool

    PH_Parameters: Optional[WufiPH_Parameters] = None
    DHW_Parameters: Optional[WufiDHW_Parameters] = None
    Heating_Parameters: Optional[WufiHeating_Parameters] = None
    Cooling_Parameters: Optional[WufiCooling_Parameters] = None
    Ventilation_Parameters: Optional[WufiVentilation_Parameters] = None

    HeatRecovery: Optional[wufi_unit._Percentage] = None
    MoistureRecovery: Optional[wufi_unit._Percentage] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiSystem(BaseModel):
    Name: str
    Type: int
    IdentNr: int
    ZonesCoverage: List[WufiZoneCoverage]
    PHDistribution: WufiPHDistribution
    Devices: List[WufiDevice]

    @validator("ZonesCoverage", allow_reuse=True, pre=True)
    def unpack_zone_coverage(cls, v):
        return v

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiSystems(BaseModel):
    Systems: List[WufiSystem]


# ------------------------------------------------------------------------------
# -- Foundations


class WufiFoundationInterface(BaseModel):
    # -- Common Attributes
    Name: Optional[str] = None
    SettingFloorSlabType: int
    FloorSlabType: int

    # -- Heated Basement
    FloorSlabArea: Optional[wufi_unit.M2] = None
    U_ValueBasementSlab: Optional[wufi_unit.Watts_per_M2K] = None
    FloorSlabPerimeter: Optional[wufi_unit.M] = None
    U_ValueBasementWall: Optional[wufi_unit.Watts_per_M2K] = None
    DepthBasementBelowGroundSurface: Optional[wufi_unit.M] = None

    # -- Unheated Basement
    HeightBasementWallAboveGrade: Optional[wufi_unit.M] = None
    FloorCeilingArea: Optional[wufi_unit.M2] = None
    U_ValueCeilingToUnheatedCellar: Optional[wufi_unit.Watts_per_M2K] = None
    U_ValueWallAboveGround: Optional[wufi_unit.Watts_per_M2K] = None
    BasementVolume: Optional[wufi_unit.M3] = None

    # -- Slab on Grade
    PositionPerimeterInsulation: Optional[int] = None
    PerimeterInsulationWidthDepth: Optional[wufi_unit.M] = None
    ConductivityPerimeterInsulation: Optional[wufi_unit.Watts_per_MK] = None
    ThicknessPerimeterInsulation: Optional[wufi_unit.M] = None

    # -- Vented Crawlspace
    U_ValueCrawlspaceFloor: Optional[wufi_unit.Watts_per_M2K] = None
    CrawlspaceVentOpenings: Optional[wufi_unit.M2] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Variants


class WufiInternalGainsAdditionalData(BaseModel):
    EvaporationHeatPerPerson: wufi_unit.Watts
    HeatLossFluschingWC: bool
    QuantityWCs: wufi_unit._Int
    RoomCategory: int
    UseDefaultValuesSchool: bool
    MarginalPerformanceRatioDHW: Optional[wufi_unit._Percentage] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiPH_Building(BaseModel):
    IdentNr: int
    BuildingCategory: int
    OccupancyTypeResidential: Optional[int] = None
    OccupancyTypeNonResidential: Optional[int] = None
    BuildingStatus: int
    BuildingType: int
    OccupancySettingMethod: int
    NumberUnits: Optional[wufi_unit._Int] = None
    CountStories: Optional[int] = None
    EnvelopeAirtightnessCoefficient: wufi_unit.M3_per_Hour_per_M2
    SummerHRVHumidityRecovery: float
    FoundationInterfaces: Optional[List[WufiFoundationInterface]] = None
    InternalGainsAdditionalData: Optional[WufiInternalGainsAdditionalData] = None
    MechanicalRoomTemperature: wufi_unit.DegreeC
    IndoorTemperature: wufi_unit.DegreeC
    OverheatingTemperatureThreshold: wufi_unit.DegreeC
    NonCombustibleMaterials: bool
    BuildingWindExposure: int = 1

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)

    @validator("BuildingWindExposure", allow_reuse=True, pre=True)
    def validate_wind_exposure(cls, v: int | None):
        if v == None:
            return 1
        return v


class WufiPassivehouseData(BaseModel):
    PH_CertificateCriteria: int
    PH_SelectionTargetData: int
    AnnualHeatingDemand: wufi_unit.kWh_per_M2
    AnnualCoolingDemand: wufi_unit.kWh_per_M2
    PeakHeatingLoad: wufi_unit.Watts_per_M2
    PeakCoolingLoad: wufi_unit.Watts_per_M2
    PH_Buildings: List[WufiPH_Building]
    UseWUFIMeanMonthShading: bool

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiPlugin(BaseModel):
    InsertPlugIn: Optional[bool] = None
    Name_dll: Optional[Any] = None
    StatusPlugIn: Optional[Any] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiVariant(BaseModel):
    IdentNr: int
    Name: Optional[str] = None
    Remarks: Optional[str] = None
    PlugIn: Optional[WufiPlugin] = None
    Graphics_3D: WufiGraphics_3D
    Building: WufiBuilding
    PassivehouseData: WufiPassivehouseData
    HVAC: WufiSystems
    ClimateLocation: WufiClimateLocation  # Be sure this comes AFTER PassiveHouseData to validate...

    @validator("HVAC", allow_reuse=True, pre=True)
    def unpack_hvac(cls, v: WufiSystems):
        return v

    @validator("ClimateLocation", pre=False)
    def check_source_energy_factors(cls, v: WufiClimateLocation, values: Dict[str, Any]):
        """Ensure the ClimateLocation's Energy and CO2 conversion factor lists are populated properly.

        If the XML file is set to one of the 'standard' factor sets (USA, Germany, etc..),
        in that case the XML file not have any values, so load those from the Standards-Library.
        """
        if not v.PH_ClimateLocation:
            return v

        passivehouse_data: WufiPassivehouseData = values["PassivehouseData"]

        if not v.PH_ClimateLocation.CO2FactorsUserDef:
            v.PH_ClimateLocation.set_standard_pe_factors(passivehouse_data.PH_CertificateCriteria)

        if not v.PH_ClimateLocation.PEFactorsUserDef:
            v.PH_ClimateLocation.set_standard_co2_factors(passivehouse_data.PH_CertificateCriteria)

        return v

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Constructions


class WufiMaterial(BaseModel):
    Name: str
    ThermalConductivity: wufi_unit.Watts_per_MK
    BulkDensity: wufi_unit.KG_per_M3
    Porosity: wufi_unit._Percentage
    HeatCapacity: wufi_unit.Joule_per_KGK
    WaterVaporResistance: wufi_unit.WUFI_Vapor_Resistance_Factor
    ReferenceWaterContent: wufi_unit.KG_per_M3

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiDivisionV(BaseModel):
    Distance: wufi_unit.M
    ExpandingContracting: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiDivisionH(BaseModel):
    Distance: wufi_unit.M
    ExpandingContracting: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiMaterialIDNr(BaseModel):
    Type: int
    IdentNr_Object: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiLayer(BaseModel):
    Thickness: wufi_unit.M
    Material: WufiMaterial
    ExchangeDivisionHorizontal: Optional[List[WufiDivisionH]] = None
    ExchangeDivisionVertical: Optional[List[WufiDivisionV]] = None
    ExchangeMaterialIdentNrs: Optional[List[WufiMaterialIDNr]] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiExchangeMaterial(BaseModel):
    IdentNr: int
    Name: str
    ThermalConductivity: wufi_unit.Watts_per_MK
    BulkDensity: wufi_unit.KG_per_M3
    HeatCapacity: wufi_unit.Joule_per_KGK

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiAssembly(BaseModel):
    IdentNr: int
    Name: str
    Order_Layers: int
    Grid_Kind: int
    Layers: List[WufiLayer]
    ExchangeMaterials: Optional[List[WufiExchangeMaterial]] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiWindowType(BaseModel):
    IdentNr: int
    Name: str
    Uw_Detailed: bool
    GlazingFrameDetailed: bool
    FrameFactor: wufi_unit._Percentage
    U_Value: wufi_unit.Watts_per_M2K
    U_Value_Glazing: wufi_unit.Watts_per_M2K
    MeanEmissivity: wufi_unit._Percentage
    g_Value: wufi_unit._Percentage
    SHGC_Hemispherical: wufi_unit._Percentage
    U_Value_Frame: wufi_unit.Watts_per_M2K

    # -- WUFI might not provide these...
    Frame_Width_Left: Optional[wufi_unit.M] = None
    Frame_Psi_Left: Optional[wufi_unit.Watts_per_MK] = None
    Frame_U_Left: Optional[wufi_unit.Watts_per_M2K] = None
    Glazing_Psi_Left: Optional[wufi_unit.Watts_per_MK] = None

    Frame_Width_Right: Optional[wufi_unit.M] = None
    Frame_Psi_Right: Optional[wufi_unit.Watts_per_MK] = None
    Frame_U_Right: Optional[wufi_unit.Watts_per_M2K] = None
    Glazing_Psi_Right: Optional[wufi_unit.Watts_per_MK] = None

    Frame_Width_Top: Optional[wufi_unit.M] = None
    Frame_Psi_Top: Optional[wufi_unit.Watts_per_MK] = None
    Frame_U_Top: Optional[wufi_unit.Watts_per_M2K] = None
    Glazing_Psi_Top: Optional[wufi_unit.Watts_per_MK] = None

    Frame_Width_Bottom: Optional[wufi_unit.M] = None
    Frame_Psi_Bottom: Optional[wufi_unit.Watts_per_MK] = None
    Frame_U_Bottom: Optional[wufi_unit.Watts_per_M2K] = None
    Glazing_Psi_Bottom: Optional[wufi_unit.Watts_per_MK] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiSolarProtectionType(BaseModel):
    IdentNr: int
    Name: str
    OperationMode: int
    MaxRedFactorRadiation: Optional[wufi_unit._Percentage] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Patterns (Schedules)


class WufiUtilizationPatternVent(BaseModel):
    Name: str
    IdentNr: int
    OperatingDays: float
    OperatingWeeks: float
    Maximum_DOS: float
    Maximum_PDF: float
    Standard_DOS: float
    Standard_PDF: float
    Basic_DOS: float
    Basic_PDF: float
    Minimum_DOS: float
    Minimum_PDF: float

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiUtilizationPattern(BaseModel):
    IdentNr: int
    Name: str
    HeightUtilizationLevel: float
    BeginUtilization: wufi_unit.Hour
    EndUtilization: wufi_unit.Hour
    AnnualUtilizationDays: wufi_unit.Days_per_Year
    IlluminationLevel: wufi_unit.Lux
    RelativeAbsenteeism: wufi_unit._Percentage
    PartUseFactorPeriodForLighting: wufi_unit._Percentage

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Project


class WufiDateProject(BaseModel):
    Year: int
    Month: int
    Day: int
    Hour: int
    Minutes: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WufiProjectData(BaseModel):
    Year_Construction: Optional[int] = None
    OwnerIsClient: Optional[bool] = None
    WhiteBackgroundPictureBuilding: Optional[bool] = None
    Customer_Name: Optional[str] = None
    Customer_Street: Optional[str] = None
    Customer_Locality: Optional[str] = None
    Customer_PostalCode: Optional[str] = None
    Customer_Tel: Optional[str] = None
    Customer_Email: Optional[str] = None
    Building_Name: Optional[str] = None
    Building_Street: Optional[str] = None
    Building_Locality: Optional[str] = None
    Building_PostalCode: Optional[str] = None
    Owner_Name: Optional[str] = None
    Owner_Street: Optional[str] = None
    Owner_Locality: Optional[str] = None
    Owner_PostalCode: Optional[str] = None
    Responsible_Name: Optional[str] = None
    Responsible_Street: Optional[str] = None
    Responsible_Locality: Optional[str] = None
    Responsible_PostalCode: Optional[str] = None
    Responsible_Tel: Optional[str] = None
    Responsible_LicenseNr: Optional[str] = None
    Responsible_Email: Optional[str] = None
    Date_Project: Optional[WufiDateProject] = None

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Root


class WUFIplusProject(BaseModel):
    DataVersion: int
    UnitSystem: int
    ProgramVersion: str
    Scope: int
    DimensionsVisualizedGeometry: int
    ProjectData: WufiProjectData
    UtilizationPatternsPH: Optional[List[WufiUtilizationPattern]] = None
    UtilisationPatternsVentilation: Optional[List[WufiUtilizationPatternVent]] = None
    WindowTypes: Optional[List[WufiWindowType]] = None
    Assemblies: Optional[List[WufiAssembly]] = None
    Variants: Optional[List[WufiVariant]] = None
    SolarProtectionTypes: Optional[List[WufiSolarProtectionType]] = None

    @validator("UnitSystem", allow_reuse=True, pre=True)
    def unpack_unit_system(cls, v) -> int:
        """
        Since the model gets converted to SI units when it is read in,
        always set this to 1, no matter what.
        """
        return 1

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Construction, Materials Classes"""

from __future__ import annotations

import uuid
import warnings
from dataclasses import dataclass, field
from typing import ClassVar, Generator, Iterable, List, Optional, Union

# -----------------------------------------------------------------------------
# Materials


@dataclass
class PhxColor:
    # Default color is white
    alpha: int = 255
    red: int = 255
    green: int = 255
    blue: int = 255


@dataclass
class PhxMaterial:
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)

    display_name: str = ""
    conductivity: float = 0.0
    density: float = 0.0
    porosity: float = 0.0
    heat_capacity: float = 0.0
    water_vapor_resistance: float = 0.0
    reference_water: float = 0.0
    percentage_of_assembly: float = 1.0
    argb_color: PhxColor = field(default_factory=PhxColor)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    def __eq__(self, other: PhxMaterial) -> bool:
        return self.id_num == other.id_num

    def __hash__(self) -> int:
        return hash(self.id_num)


# -----------------------------------------------------------------------------
# Layers (and Mixed Materials)


@dataclass
class PhxLayerDivisionCell:
    """A single cell aty a column/row in a PhxLayerDivisionGrid, with a specific material."""

    row: int
    column: int
    material: PhxMaterial
    expanding_contracting: int = 2  # 2="Exp./Contr."


@dataclass
class PhxLayerDivisionGrid:
    """A grid of PhxLayerDivisionCells to support 'mixed' materials.

    The Cell grid is ordered from top-left to bottom-right:

    |    | C0  | C1  | C2  | ...
    |:---|:---:|:---:|:---:|:---:
    | R0 | 0,0 | 1,0 | 2,0 | ...
    | R1 | 0,1 | 1,1 | 2,1 | ...
    | R2 | 0,2 | 1,2 | 2,2 | ...
    """

    _row_heights: List[float] = field(default_factory=list)
    _column_widths: List[float] = field(default_factory=list)
    _cells: List[PhxLayerDivisionCell] = field(default_factory=list)

    @property
    def column_widths(self) -> List[float]:
        """Return the list of column widths."""
        return self._column_widths

    @property
    def column_count(self) -> int:
        """Return the number of columns in the grid."""
        return len(self._column_widths)

    @property
    def row_heights(self) -> List[float]:
        """Return the list of row heights."""
        return self._row_heights

    @property
    def row_count(self) -> int:
        """Return the number of rows in the grid."""
        return len(self._row_heights)

    @property
    def cell_count(self) -> int:
        """Return the total number of cells in the grid."""
        return self.row_count * self.column_count

    def set_column_widths(self, _column_widths: Iterable[float]) -> None:
        """Set the column widths of the grid."""
        self._column_widths = []
        for width in _column_widths:
            self.add_new_column(width)

    def add_new_column(self, _column_width: float) -> None:
        """Add a new COLUMN to the grid with the given width."""
        self._column_widths.append(float(_column_width))

    def set_row_heights(self, _row_heights: Iterable[float]) -> None:
        """Set the row heights of the grid."""
        self._row_heights = []
        for height in _row_heights:
            self.add_new_row(height)

    def add_new_row(self, _row_height: float) -> None:
        """Add a new ROW to the grid with the given height. Will add a default column if none are set."""
        self._row_heights.append(float(_row_height))

    def get_cell(self, _column: int, _row: int) -> Optional[PhxLayerDivisionCell]:
        """Get the PhxLayerDivisionCell at the given column and row position."""
        for cell in self._cells:
            if cell.column == _column and cell.row == _row:
                return cell
        return None

    def set_cell_material(self, _column_num: int, _row_num: int, _phx_material: PhxMaterial) -> None:
        """Set the PhxMaterial for a specific cell in the grid by its column/row position.

        Cells are indexed by their column and row position stating from top-left:

        |   | C0   | C1  | C2  | ...
        |---|------|-----|-----|----
        |R0 | 0,0  | 1,0 | 2,0 | ...
        |R1 | 0,1  | 1,1 | 2,1 | ...
        |R2 | 0,2  | 1,2 | 2,2 | ...
        """
        if _column_num >= self.column_count:
            raise IndexError(
                f"Column number '{_column_num}' is out of range."
                "Please set the columns before assigning division materials."
            )

        if _row_num >= self.row_count:
            raise IndexError(
                f"Row number '{_row_num}' is out of range." "Please set the rows before assigning division materials."
            )

        # -- See if the cell already exists, if so reset its material
        # -- if not, create a new cell.
        cell = self.get_cell(_column_num, _row_num)
        if cell:
            cell.material = _phx_material
        else:
            cell = PhxLayerDivisionCell(row=_row_num, column=_column_num, material=_phx_material)

        self._cells.append(cell)

    def get_cell_material(self, _column_num: int, _row_num: int) -> Optional[PhxMaterial]:
        """Get the PhxMaterial for a specific cell in the grid by its column/row position."""
        for cell in self._cells:
            if cell.row == _row_num and cell.column == _column_num:
                return cell.material
        return None

    def get_cell_area(self, _column_num: int, _row_num) -> float:
        """Get the area of the cell (row-height * column-width)."""
        col_width = self.column_widths[_column_num]
        row_height = self.row_heights[_row_num]
        return col_width * row_height


@dataclass
class PhxLayer:
    """A single layer in a PhxConstructionOpaque.

    This layer may be a single material or a grid of multiple materials.
    """

    thickness_m: float = 0.0
    divisions: PhxLayerDivisionGrid = field(default_factory=PhxLayerDivisionGrid)
    _material: PhxMaterial = field(default_factory=PhxMaterial)

    def _update_material_percentages(self) -> None:
        """Update the percentage of the assembly for each material in the layer."""
        if self.divisions.cell_count == 0:
            return None

        material_areas = {}
        for row in range(self.divisions.row_count or 1):
            for column in range(self.divisions.column_count or 1):
                # -- Find the cell's area and material
                cell_area = self.divisions.get_cell_area(column, row)
                cell_material = self.divisions.get_cell_material(column, row) or self.material

                # -- Update the material area record
                if id(cell_material) not in material_areas:
                    record = {"material": cell_material, "area": cell_area}
                    material_areas[id(cell_material)] = record
                else:
                    material_areas[id(cell_material)]["area"] += cell_area

        # -- Re-set the percentage of the assembly for each material
        total_area = sum([record["area"] for record in material_areas.values()])
        for record in material_areas.values():
            record["material"].percentage_of_assembly = record["area"] / total_area

        return None

    @property
    def materials(self) -> List[PhxMaterial]:
        """Return a list of all the PhxMaterials in the Layer."""
        self._update_material_percentages()
        return [self.material] + self.exchange_materials

    @materials.setter
    def materials(self, _: PhxMaterial) -> None:
        msg = (
            "The 'materials' property is deprecated. Please use the 'set_material()' method or "
            "the 'divisions' attribute to setup mixed materials."
        )
        raise DeprecationWarning(msg)

    @property
    def material(self) -> PhxMaterial:
        """Return the base PhxMaterial of the Layer."""
        return self._material

    def add_material(self, _material: PhxMaterial) -> None:
        """Add a PhxMaterial to the self.materials collection."""
        self._material = _material
        msg = (
            "The 'add_material' method is deprecated. Please use the 'set_material()' method or "
            "the 'divisions' attribute for mixed materials."
        )
        warnings.warn(msg, DeprecationWarning)

    def set_material(self, _material: PhxMaterial) -> None:
        """Set the PhxMaterial for the layer."""
        self._material = _material

    @property
    def thickness_mm(self) -> float:
        """Returns the thickness of the layer in MM"""
        return self.thickness_m * 1000

    @classmethod
    def from_total_u_value(cls, _total_u_value: float) -> PhxLayer:
        """Returns a new PhxLayer with a single material with the given U-Value.
        note that this will assign a default thickness of 1m to the layer.
        """
        obj = cls()
        obj.thickness_m = 1

        new_mat = PhxMaterial()
        new_mat.conductivity = obj.thickness_m * _total_u_value
        new_mat.display_name = "Material"
        obj.set_material(new_mat)

        return obj

    @property
    def layer_resistance(self) -> float:
        """Returns the thermal-resistance of the layer in m2K/W"""
        try:
            return (1 / self.material.conductivity) * self.thickness_m
        except ZeroDivisionError:
            return 0.0

    @property
    def layer_conductance(self) -> float:
        """Returns the thermal-conductance of the layer in W/m2K"""
        try:
            return 1 / self.layer_resistance
        except ZeroDivisionError:
            return 0.0

    @property
    def division_materials(self) -> list[PhxMaterial]:
        """Returns a list of all the PhxLayerDivisionCell PhxMaterials ordered by column and then row:

        |    | C0  | C1  | C2  | ...
        |:---|:---:|:---:|:---:|:---:
        | R0 | 0   | 4   | 8   | ...
        | R1 | 1   | 5   | 9   | ...
        | R2 | 2   | 6   | 10  | ...
        | R3 | 3   | 7   | 11  | ...
        """
        if self.divisions.row_count == 0 and self.divisions.column_count == 0:
            return []

        return [
            self.divisions.get_cell_material(column, row) or self.material
            for column in range(self.divisions.column_count or 1)
            for row in range(self.divisions.row_count or 1)
        ]

    @property
    def exchange_materials(self) -> List[PhxMaterial]:
        """Returns a list of all the 'Exchange' materials (for mixed layers) in all the Division Cells."""
        return list({m for m in self.division_materials if m != self.material})

    @property
    def division_material_id_numbers(self) -> List[int]:
        """Returns a list of all the 'Exchange' material id-numbers in all the Division Cells.

        Will return -1 if the cell has the Layer's Material.
        """
        id_numbers_ = []
        for m in self.division_materials:
            if m != self.material:
                id_numbers_.append(m.id_num)
            else:
                id_numbers_.append(-1)
        return id_numbers_


# -----------------------------------------------------------------------------
# Construction


@dataclass
class PhxConstructionOpaque:
    _count: ClassVar[int] = 0

    _identifier: Union[uuid.UUID, str] = field(init=False, default_factory=uuid.uuid4)
    id_num: int = field(init=False, default=0)
    display_name: str = ""
    layer_order: int = 2  # Outside to Inside
    grid_kind: int = 2  # Medium
    layers: list[PhxLayer] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def identifier(self) -> str:
        return str(self._identifier)

    @identifier.setter
    def identifier(self, _in: Optional[str]) -> None:
        if not _in:
            return
        self._identifier = str(_in)

    @property
    def r_value(self) -> float:
        return sum(l.layer_resistance for l in self.layers)

    @property
    def u_value(self) -> float:
        try:
            return 1 / self.r_value
        except ZeroDivisionError:
            return 0.0

    def __hash__(self) -> int:
        return hash(self.identifier)

    @classmethod
    def from_total_u_value(cls, _total_u_value: float, _display_name: str = "") -> PhxConstructionOpaque:
        """Returns a new PhxConstructionOpaque with a single layer with the given U-Value."""
        obj = cls()
        obj.display_name = _display_name or f"U-Value: {_total_u_value}"
        obj.layers.append(PhxLayer.from_total_u_value(_total_u_value))
        return obj

    @property
    def exchange_materials(self) -> List[PhxMaterial]:
        """Returns a flat list of all the 'Exchange' materials (for mixed layers) from all the Layers."""
        return [material for layer in self.layers for material in layer.exchange_materials]


# -----------------------------------------------------------------------------
# Windows


@dataclass
class PhxWindowFrameElement:
    width: float = 0.1  # m
    u_value: float = 1.0  # W/m2k
    psi_glazing: float = 0.00  # W/mk
    psi_install: float = 0.00  # W/mk

    @classmethod
    def from_total_u_value(cls, _total_u_value: float) -> PhxWindowFrameElement:
        """Returns a new PhxWindowFrameElement with u-values set to a single value."""
        obj = cls()
        obj.width = 0.1  # m
        obj.u_value = _total_u_value
        obj.psi_glazing = 0.0
        obj.psi_install = 0.0
        return obj


@dataclass
class PhxConstructionWindow:
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    _identifier: Union[uuid.UUID, str] = field(init=False, default_factory=uuid.uuid4)
    display_name: str = ""
    _glazing_type_display_name: str = ""
    _frame_type_display_name: str = ""

    use_detailed_uw: bool = True
    use_detailed_frame: bool = True

    u_value_window: float = 1.0
    u_value_glass: float = 1.0
    u_value_frame: float = 1.0

    frame_top: PhxWindowFrameElement = field(default_factory=PhxWindowFrameElement)
    frame_right: PhxWindowFrameElement = field(default_factory=PhxWindowFrameElement)
    frame_bottom: PhxWindowFrameElement = field(default_factory=PhxWindowFrameElement)
    frame_left: PhxWindowFrameElement = field(default_factory=PhxWindowFrameElement)
    frame_factor: float = 0.75

    glass_mean_emissivity: float = 0.1
    glass_g_value: float = 0.4

    _id_num_shade: int = -1

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def glazing_type_display_name(self) -> str:
        return self._glazing_type_display_name or self.display_name

    @glazing_type_display_name.setter
    def glazing_type_display_name(self, _in: str) -> None:
        self._glazing_type_display_name = _in

    @property
    def frame_type_display_name(self) -> str:
        return self._frame_type_display_name or self.display_name

    @frame_type_display_name.setter
    def frame_type_display_name(self, _in: str) -> None:
        self._frame_type_display_name = _in

    @property
    def identifier(self) -> str:
        return str(self._identifier)

    @identifier.setter
    def identifier(self, _in: str) -> None:
        if not _in:
            return
        self._identifier = str(_in)

    @property
    def id_num_shade(self) -> int:
        return self._id_num_shade

    @id_num_shade.setter
    def id_num_shade(self, _in: Optional[int]) -> None:
        if _in != None:
            self._id_num_shade = _in

    @classmethod
    def from_total_u_value(
        cls, _total_u_value: float, _g_value: float = 0.4, _display_name: str = ""
    ) -> PhxConstructionWindow:
        """Returns a new PhxConstructionWindow with all u-values set to a single value."""
        obj = cls()
        obj.display_name = _display_name or f"U-Value: {_total_u_value}"
        obj.frame_type_display_name = _display_name or f"U-Value: {_total_u_value}"
        obj.glazing_type_display_name = _display_name or f"U-Value: {_total_u_value}"

        obj.u_value_window = _total_u_value
        obj.u_value_glass = _total_u_value
        obj.u_value_frame = _total_u_value

        obj.glass_g_value = _g_value

        obj.frame_top = PhxWindowFrameElement.from_total_u_value(_total_u_value)
        obj.frame_right = PhxWindowFrameElement.from_total_u_value(_total_u_value)
        obj.frame_bottom = PhxWindowFrameElement.from_total_u_value(_total_u_value)
        obj.frame_left = PhxWindowFrameElement.from_total_u_value(_total_u_value)

        return obj

    @property
    def frames(self) -> Generator[PhxWindowFrameElement, None, None]:
        """Returns a generator of all frame elements in order: Top, Right, Bottom, Left."""
        for frame in (
            self.frame_top,
            self.frame_right,
            self.frame_bottom,
            self.frame_left,
        ):
            yield frame

    def set_all_frames_u_value(self, _u_value: float) -> None:
        """Sets the u-value of all frame elements to the given value."""
        for frame in self.frames:
            frame.u_value = _u_value

    def set_all_frames_width(self, _width: float) -> None:
        """Sets the width of all frame elements to the given value."""
        for frame in self.frames:
            frame.width = _width

    def set_all_frames_psi_glazing(self, _psi_glazing: float) -> None:
        """Sets the psi_glazing of all frame elements to the given value."""
        for frame in self.frames:
            frame.psi_glazing = _psi_glazing

    def set_all_frames_psi_install(self, _psi_install: float) -> None:
        """Sets the psi_install of all frame elements to the given value."""
        for frame in self.frames:
            frame.psi_install = _psi_install



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Model class for a PHPP DHW Piping Elements."""

from dataclasses import dataclass
from functools import partial
from typing import List

from PHX.model.hvac import piping
from PHX.PHPP.phpp_localization import shape_model
from PHX.xl import xl_data


@dataclass
class RecircPipingInput:
    """Model class for a single DHW Recirculation Pipe Element."""

    __slots__ = ("shape", "phx_pipe", "pipe_group_num")
    shape: shape_model.Dhw
    phx_pipe: List[piping.PhxPipeSegment]
    pipe_group_num: int

    @property
    def input_column(self) -> str:
        """Return the right input column based on the pipe-group-number."""
        return xl_data.col_offset(self.shape.recirc_piping.input_col_start, self.pipe_group_num)

    def _create_range(
        self,
        _field_name: str,
        _row_num: int,
    ) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        return f"{self.input_column}{_row_num + getattr(self.shape.recirc_piping.input_rows_offset, _field_name).row}"

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.recirc_piping.input_rows_offset, _field_name).unit

    def _bool_as_x(self, _value: bool):
        if _value == True:
            return "x"
        else:
            return ""

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemDHW = partial(xl_data.XlItem, _sheet_name)
        total_length = sum(s.length_m for s in self.phx_pipe)

        return [
            XLItemDHW(
                create_range("total_length"),
                total_length,
                "M",
                self._get_target_unit("total_length"),
            ),
            XLItemDHW(
                create_range("diameter"),
                sum(s.diameter_m * s.length_m for s in self.phx_pipe) / total_length,
                "M",
                self._get_target_unit("diameter"),
            ),
            XLItemDHW(
                create_range("insul_thickness"),
                sum(s.insulation_thickness_m * s.length_m for s in self.phx_pipe) / total_length,
                "M",
                self._get_target_unit("insul_thickness"),
            ),
            XLItemDHW(
                create_range("insul_reflective"),
                self._bool_as_x(any([s.insulation_reflective for s in self.phx_pipe])),
            ),
            XLItemDHW(
                create_range("insul_conductivity"),
                sum(s.insulation_conductivity * s.length_m for s in self.phx_pipe) / total_length,
                "W/MK",
                self._get_target_unit("insul_conductivity"),
            ),
            XLItemDHW(
                create_range("daily_period"),
                sum(s.daily_period * s.length_m for s in self.phx_pipe) / total_length,
                "",
                self._get_target_unit("daily_period"),
            ),
            XLItemDHW(
                create_range("water_temp"),
                sum(s.water_temp_c * s.length_m for s in self.phx_pipe) / total_length,
                "C",
                self._get_target_unit("water_temp"),
            ),
        ]


@dataclass
class BranchPipingInput:
    """Model class for a single DHW Branch Pipe Element."""

    __slots__ = ("shape", "phx_pipe", "pipe_group_num", "num_tap_points")
    shape: shape_model.Dhw
    phx_pipe: List[piping.PhxPipeSegment]
    pipe_group_num: int
    num_tap_points: int

    @property
    def input_column(self) -> str:
        """Return the right input column based on the pipe-group-number."""
        return xl_data.col_offset(self.shape.branch_piping.input_col_start, self.pipe_group_num)

    def create_range(self, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        return f"{self.input_column}{_row_num}"

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.branch_piping.input_rows_offset, _field_name).unit

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of Branch Piping Xl-Write items."""

        XLItemDHW = partial(xl_data.XlItem, _sheet_name)
        total_length = sum(s.length_m for s in self.phx_pipe)
        if total_length == 0:
            return []
        return [
            # -- Branch Piping
            XLItemDHW(
                self.create_range(_row_num + self.shape.branch_piping.input_rows_offset.water_temp.row),  # type: ignore
                sum(s.water_temp_c * s.length_m for s in self.phx_pipe) / total_length,
                "C",
                self._get_target_unit("water_temp"),
            ),
            XLItemDHW(
                self.create_range(_row_num + self.shape.branch_piping.input_rows_offset.diameter.row),  # type: ignore
                sum(s.diameter_m * s.length_m for s in self.phx_pipe) / total_length,
                "M",
                self._get_target_unit("diameter"),
            ),
            XLItemDHW(
                self.create_range(_row_num + self.shape.branch_piping.input_rows_offset.total_length.row),  # type: ignore
                total_length,
                "M",
                self._get_target_unit("total_length"),
            ),
            # -- Add Tapping Points
            XLItemDHW(
                self.create_range(_row_num + self.shape.branch_piping.input_rows_offset.num_taps),
                self.num_tap_points,
            ),
        ]



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Model class for a PHPP Areas / Surface-Entry row"""

import re
from dataclasses import dataclass
from functools import partial
from typing import Dict, List, Optional

from PHX.model import components, geometry
from PHX.model.enums.building import ComponentExposureExterior, ComponentFaceType
from PHX.PHPP.phpp_localization import shape_model
from PHX.PHPP.phpp_model import version
from PHX.xl import xl_data


@dataclass
class SurfaceRow:
    """A single Areas/Surface entry row."""

    __slots__ = ("shape", "phx_polygon", "phx_component", "phpp_assembly_id_name", "phpp_version")
    shape: shape_model.Areas
    phx_polygon: geometry.PhxPolygon
    phx_component: components.PhxComponentOpaque
    phpp_assembly_id_name: Optional[str]
    phpp_version: version.PHPPVersion

    @property
    def phpp_group_number_format(self) -> str:
        """Return the correct PHPP 'Group Format' depending on the PHPP version."""
        if self.phpp_version.number_major == "9":
            return "{}"
        else:
            return "{}-"

    @property
    def phpp_group_number(self) -> str:
        """Return the correct PHPP 'Group Number' depending on the exposure and type."""

        if self.phx_component.exposure_exterior == ComponentExposureExterior.SURFACE:
            return self.phpp_group_number_format.format(18)
        elif self.phx_component.face_type == ComponentFaceType.WALL:
            if self.phx_component.exposure_exterior == ComponentExposureExterior.EXTERIOR:
                return self.phpp_group_number_format.format(8)
            else:
                return self.phpp_group_number_format.format(9)
        elif self.phx_component.face_type == ComponentFaceType.FLOOR:
            return self.phpp_group_number_format.format(11)
        elif self.phx_component.face_type == ComponentFaceType.ROOF_CEILING:
            return self.phpp_group_number_format.format(10)
        else:
            return self.phpp_group_number_format.format(12)

    def _create_range(self, _field_name: str, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        return f"{getattr(self.shape.surface_rows.inputs, _field_name).column}{_row_num}"

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.surface_rows.inputs, _field_name).unit

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this Surface Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemAreas = partial(xl_data.XlItem, _sheet_name)
        return [
            XLItemAreas(create_range("description"), f"'{self.phx_polygon.display_name}"),
            # -- Note, Add a "-" to the group number so that it is 'text' otherwise the
            # -- IP PHPP won't recognize it properly.
            XLItemAreas(create_range("group_number"), self.phpp_group_number),
            XLItemAreas(create_range("quantity"), 1),
            XLItemAreas(
                create_range("area"),
                self.phx_polygon.area,
                "M2",
                self._get_target_unit("area"),
            ),
            XLItemAreas(create_range("assembly_id"), self.phpp_assembly_id_name),
            XLItemAreas(create_range("orientation"), self.phx_polygon.cardinal_orientation_angle),
            XLItemAreas(create_range("angle"), self.phx_polygon.angle_from_horizontal),
            XLItemAreas(create_range("shading"), 0.5),
            XLItemAreas(create_range("absorptivity"), 0.6),
            XLItemAreas(create_range("emissivity"), 0.9),
        ]


def get_name_from_assembly_id(
    _phpp_assembly_id: Optional[xl_data.xl_range_single_value],
) -> str:
    """Return the face's construction PHPP-Name (ie: "MyConst") from id-string."""
    try:
        return str(_phpp_assembly_id).split("-", 1)[1]
    except:
        if _phpp_assembly_id == None or _phpp_assembly_id == "None":
            return ""
        else:
            msg = f"Error getting construction PHPP-Name? " f"Could not split {_phpp_assembly_id} on '-'?"
            raise Exception(msg)


@dataclass
class ExistingSurfaceRow:
    """The data from an existing PHPP Surface Entry row."""

    __slots__ = ("shape", "data", "group_type_exposure_map")
    shape: shape_model.Areas
    data: list
    group_type_exposure_map: Dict[int, str]  # From io_areas.Areas() parent class

    @property
    def is_empty(self) -> bool:
        if self.name in ["-", "", "None", None]:
            return True
        if self.face_group_type_phpp_string in ["-", "", "None", None]:
            return True
        if self.face_group_type_phpp_number == 0:
            return True
        if self.face_type == ComponentFaceType.NONE:
            return True
        if self.face_exposure == ComponentExposureExterior.NONE:
            return True
        return False

    @property
    def name(self) -> str:
        col_letter = self.shape.surface_rows.inputs.description.column
        col_number_as_index = xl_data.xl_ord(str(col_letter)) - 65
        return str(self.data[col_number_as_index])

    @property
    def face_group_type_phpp_string(self) -> str:
        """Return the face's Group-Type string from the row-data."""
        col_letter = self.shape.surface_rows.inputs.group_number.column
        col_number_as_index = xl_data.xl_ord(str(col_letter)) - 65
        return str(self.data[col_number_as_index])

    @property
    def face_group_type_phpp_number(self) -> int:
        """Return the face's Group-Type number as an int from the Group-Type string."""
        phpp_data = self.face_group_type_phpp_string
        if phpp_data in ["-", "", "None", None]:
            return 0

        result = re.split(r"\D+", phpp_data, 2)
        if not result:
            msg = (
                f"Error getting Group-Type number? Could not find a number "
                f"in the Group-Type string from PHPP: '{phpp_data}'?"
            )
            raise Exception(msg)

        try:
            return int(result[0])
        except:
            msg = (
                f"Error getting Group-Type number? Could not convert '{result[0]}' "
                f"to int from the PHPP value: '{phpp_data}' or type: {type(phpp_data)}?"
            )
            raise Exception(msg)

    @property
    def face_exposure_phpp_letter(self) -> str:
        """Return the face's exposure-type letter ("A", "B", etc..) based on the group_type_phpp_number"""
        return self.group_type_exposure_map[self.face_group_type_phpp_number]

    @property
    def face_type(self) -> ComponentFaceType:
        """Return the face type enum (WALL, FLOOR, etc..) based on the group_type_phpp_number."""
        type_map = {
            0: ComponentFaceType.NONE,
            2: ComponentFaceType.WINDOW,
            3: ComponentFaceType.WINDOW,
            4: ComponentFaceType.WINDOW,
            5: ComponentFaceType.WINDOW,
            6: ComponentFaceType.WINDOW,
            7: ComponentFaceType.WINDOW,
            8: ComponentFaceType.WALL,
            9: ComponentFaceType.WALL,
            10: ComponentFaceType.ROOF_CEILING,
            11: ComponentFaceType.FLOOR,
            12: ComponentFaceType.CUSTOM,
            13: ComponentFaceType.CUSTOM,
            14: ComponentFaceType.CUSTOM,
            18: ComponentFaceType.ADIABATIC,
        }
        return type_map[self.face_group_type_phpp_number]

    @property
    def face_exposure(self) -> ComponentExposureExterior:
        """Return the exposure type enum (EXTERIOR, GROUND, SURFACE) based on the face_exposure_phpp_letter"""
        type_map = {
            "A": ComponentExposureExterior.EXTERIOR,
            "B": ComponentExposureExterior.GROUND,
            "": ComponentExposureExterior.NONE,
        }
        return type_map.get(self.face_exposure_phpp_letter, ComponentExposureExterior.SURFACE)

    @property
    def face_construction_phpp_id(self) -> str:
        """Return the face's construction PHPP-ID (ie: "01ud-MyConst") from the row-data."""
        col_letter = self.shape.surface_rows.inputs.assembly_id.column
        col_number_as_index = xl_data.xl_ord(str(col_letter)) - 65
        return str(self.data[col_number_as_index])

    @property
    def face_construction_phpp_name(self) -> str:
        """Return the face's construction PHPP-Name (ie: "MyConst") from the row-data."""
        return get_name_from_assembly_id(self.face_construction_phpp_id)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Site (Location and Climate) Dataclasses"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Generator, List, Optional, Tuple, Union

from PHX.model.enums.phx_site import SiteClimateSelection, SiteEnergyFactorSelection, SiteSelection


@dataclass
class PhxGround:
    ground_thermal_conductivity: float = 2.0
    ground_heat_capacity: float = 1000.0
    ground_density: float = 2000.0
    depth_groundwater: float = 3.0
    flow_rate_groundwater: float = 0.05

    def __eq__(self, other: PhxGround) -> bool:
        TOLERANCE = 0.0001
        return (
            abs(self.ground_thermal_conductivity - other.ground_thermal_conductivity) < TOLERANCE
            and abs(self.ground_heat_capacity - other.ground_heat_capacity) < TOLERANCE
            and abs(self.ground_density - other.ground_density) < TOLERANCE
            and abs(self.depth_groundwater - other.depth_groundwater) < TOLERANCE
            and abs(self.flow_rate_groundwater - other.flow_rate_groundwater) < TOLERANCE
        )


@dataclass
class PhxPEFactor:
    """Conversion Factors for Site-Energy->Primary-Energy"""

    value: float = 0.0
    unit: str = ""
    fuel_name: str = ""

    def __eq__(self, other: PhxPEFactor) -> bool:
        TOLERANCE = 0.0001
        return (
            abs(self.value - other.value) < TOLERANCE and self.unit == other.unit and self.fuel_name == other.fuel_name
        )


@dataclass
class PhxCO2Factor:
    """Conversion Factors for Site->CO2"""

    value: float = 0.0
    unit: str = ""
    fuel_name: str = ""

    def __eq__(self, other: PhxCO2Factor) -> bool:
        TOLERANCE = 0.0001
        return (
            abs(self.value - other.value) < TOLERANCE and self.unit == other.unit and self.fuel_name == other.fuel_name
        )


PhxEnergyFactorAlias = Union[PhxPEFactor, PhxCO2Factor]


@dataclass
class PhxSiteEnergyFactors:
    selection_pe_co2_factor: SiteEnergyFactorSelection = SiteEnergyFactorSelection.USER_DEFINED
    pe_factors: dict[str, PhxEnergyFactorAlias] = field(default_factory=dict)
    co2_factors: dict[str, PhxEnergyFactorAlias] = field(default_factory=dict)

    def __post_init__(self):
        self.pe_factors = {
            "OIL": PhxPEFactor(1.1, "kWh/kWh", "OIL"),
            "NATURAL_GAS": PhxPEFactor(1.1, "kWh/kWh", "NATURAL_GAS"),
            "LPG": PhxPEFactor(1.1, "kWh/kWh", "LPG"),
            "HARD_COAL": PhxPEFactor(1.1, "kWh/kWh", "HARD_COAL"),
            "WOOD": PhxPEFactor(0.2, "kWh/kWh", "WOOD"),
            "ELECTRICITY_MIX": PhxPEFactor(1.8, "kWh/kWh", "ELECTRICITY_MIX"),
            "ELECTRICITY_PV": PhxPEFactor(0.7, "kWh/kWh", "ELECTRICITY_PV"),
            "HARD_COAL_CGS_70_CHP": PhxPEFactor(0.8, "kWh/kWh", "HARD_COAL_CGS_70_CHP"),
            "HARD_COAL_CGS_35_CHP": PhxPEFactor(1.1, "kWh/kWh", "HARD_COAL_CGS_35_CHP"),
            "HARD_COAL_CGS_0_CHP": PhxPEFactor(1.5, "kWh/kWh", "HARD_COAL_CGS_0_CHP"),
            "GAS_CGS_70_CHP": PhxPEFactor(0.7, "kWh/kWh", "GAS_CGS_70_CHP"),
            "GAS_CGS_35_CHP": PhxPEFactor(1.1, "kWh/kWh", "GAS_CGS_35_CHP"),
            "GAS_CGS_0_CHP": PhxPEFactor(1.5, "kWh/kWh", "GAS_CGS_0_CHP"),
            "OIL_CGS_70_CHP": PhxPEFactor(0.8, "kWh/kWh", "OIL_CGS_70_CHP"),
            "OIL_CGS_35_CHP": PhxPEFactor(1.1, "kWh/kWh", "OIL_CGS_35_CHP"),
            "OIL_CGS_0_CHP": PhxPEFactor(1.5, "kWh/kWh", "OIL_CGS_0_CHP"),
        }
        self.co2_factors: dict[str, PhxEnergyFactorAlias] = {
            "OIL": PhxCO2Factor(309.9966, "g/kWh", "OIL"),
            "NATURAL_GAS": PhxCO2Factor(250.0171, "g/kWh", "NATURAL_GAS"),
            "LPG": PhxCO2Factor(270.0102, "g/kWh", "LPG"),
            "HARD_COAL": PhxCO2Factor(439.9864, "g/kWh", "HARD_COAL"),
            "WOOD": PhxCO2Factor(53.4289, "g/kWh", "WOOD"),
            "ELECTRICITY_MIX": PhxCO2Factor(680.0068, "g/kWh", "ELECTRICITY_MIX"),
            "ELECTRICITY_PV": PhxCO2Factor(250.0171, "g/kWh", "ELECTRICITY_PV"),
            "HARD_COAL_CGS_70_CHP": PhxCO2Factor(239.9864, "g/kWh", "HARD_COAL_CGS_70_CHP"),
            "HARD_COAL_CGS_35_CHP": PhxCO2Factor(319.9932, "g/kWh", "HARD_COAL_CGS_35_CHP"),
            "HARD_COAL_CGS_0_CHP": PhxCO2Factor(409.9966, "g/kWh", "HARD_COAL_CGS_0_CHP"),
            "GAS_CGS_70_CHP": PhxCO2Factor(-70.0102, "g/kWh", "GAS_CGS_70_CHP"),
            "GAS_CGS_35_CHP": PhxCO2Factor(129.9898, "g/kWh", "GAS_CGS_35_CHP"),
            "GAS_CGS_0_CHP": PhxCO2Factor(319.9932, "g/kWh", "GAS_CGS_0_CHP"),
            "OIL_CGS_70_CHP": PhxCO2Factor(100, "g/kWh", "OIL_CGS_70_CHP"),
            "OIL_CGS_35_CHP": PhxCO2Factor(250.0171, "g/kWh", "OIL_CGS_35_CHP"),
            "OIL_CGS_0_CHP": PhxCO2Factor(409.9966, "g/kWh", "OIL_CGS_0_CHP"),
        }

    def __eq__(self, other: PhxSiteEnergyFactors) -> bool:
        return (
            self.selection_pe_co2_factor == other.selection_pe_co2_factor
            and self.pe_factors == other.pe_factors
            and self.co2_factors == other.co2_factors
        )


@dataclass
class PhxLocation:
    """The physical location of the building."""

    latitude: float = 40.6
    longitude: float = -73.8
    site_elevation: Optional[float] = None
    climate_zone: int = 1
    hours_from_UTC: int = -4

    def __eq__(self, other: PhxLocation) -> bool:
        TOLERANCE = 0.001
        return (
            abs(self.latitude - other.latitude) < TOLERANCE
            and abs(self.longitude - other.longitude) < TOLERANCE
            and abs((self.site_elevation or 0.0) - (other.site_elevation or 0.0)) < TOLERANCE
            and self.climate_zone == other.climate_zone
            and self.hours_from_UTC == other.hours_from_UTC
        )


@dataclass
class PhxClimatePeakLoad:
    temperature_air: Optional[float] = 0.0
    radiation_north: Optional[float] = 0.0
    radiation_east: Optional[float] = 0.0
    radiation_south: Optional[float] = 0.0
    radiation_west: Optional[float] = 0.0
    radiation_global: Optional[float] = 0.0

    def __eq__(self, other: PhxClimatePeakLoad) -> bool:
        TOLERANCE = 0.001
        if self.temperature_air and other.temperature_air:
            if abs(self.temperature_air - other.temperature_air) > TOLERANCE:
                return False
        if self.radiation_north and other.radiation_north:
            if abs(self.radiation_north - other.radiation_north) > TOLERANCE:
                return False
        if self.radiation_east and other.radiation_east:
            if abs(self.radiation_east - other.radiation_east) > TOLERANCE:
                return False
        if self.radiation_south and other.radiation_south:
            if abs(self.radiation_south - other.radiation_south) > TOLERANCE:
                return False
        if self.radiation_west and other.radiation_west:
            if abs(self.radiation_west - other.radiation_west) > TOLERANCE:
                return False
        if self.radiation_global and other.radiation_global:
            if abs(self.radiation_global - other.radiation_global) > TOLERANCE:
                return False
        return True


@dataclass
class PhxClimateIterOutput:
    """Wrapper class for organizing output of the 'PhxClimate.monthly_values' property."""

    month_name: str
    month_hours: float
    temperature_air: float
    temperature_dewpoint: float
    temperature_sky: float
    radiation_north: float
    radiation_east: float
    radiation_south: float
    radiation_west: float
    radiation_global: float


@dataclass
class PhxClimate:
    """Monthly Climate Date for the building location."""

    station_elevation: float = 3.0
    selection: SiteClimateSelection = SiteClimateSelection.USER_DEFINED
    daily_temp_swing: float = 8.0
    avg_wind_speed: float = 4.0

    temperature_air: list[float] = field(default_factory=list)
    temperature_dewpoint: list[float] = field(default_factory=list)
    temperature_sky: list[float] = field(default_factory=list)

    radiation_north: list[float] = field(default_factory=list)
    radiation_east: list[float] = field(default_factory=list)
    radiation_south: list[float] = field(default_factory=list)
    radiation_west: list[float] = field(default_factory=list)
    radiation_global: list[float] = field(default_factory=list)

    peak_heating_1: PhxClimatePeakLoad = field(default_factory=PhxClimatePeakLoad)
    peak_heating_2: PhxClimatePeakLoad = field(default_factory=PhxClimatePeakLoad)
    peak_cooling_1: PhxClimatePeakLoad = field(default_factory=PhxClimatePeakLoad)
    peak_cooling_2: PhxClimatePeakLoad = field(default_factory=PhxClimatePeakLoad)

    @staticmethod
    def _list_eq(list_a: List[float], list_b: List[float]) -> bool:
        TOLERANCE = 0.001
        for a, b in zip(list_a, list_b):
            if abs(a - b) > TOLERANCE:
                return False
        return True

    def __eq__(self, other) -> bool:
        TOLERANCE = 0.01
        return (
            abs(self.station_elevation - other.station_elevation) < TOLERANCE
            and self.selection == other.selection
            and abs(self.daily_temp_swing - other.daily_temp_swing) < TOLERANCE
            and abs(self.avg_wind_speed - other.avg_wind_speed) < TOLERANCE
            and self._list_eq(self.temperature_air, other.temperature_air)
            and self._list_eq(self.temperature_dewpoint, other.temperature_dewpoint)
            and self._list_eq(self.temperature_sky, other.temperature_sky)
            and self._list_eq(self.radiation_north, other.radiation_north)
            and self._list_eq(self.radiation_east, other.radiation_east)
            and self._list_eq(self.radiation_south, other.radiation_south)
            and self._list_eq(self.radiation_west, other.radiation_west)
            and self._list_eq(self.radiation_global, other.radiation_global)
            and self.peak_heating_1 == other.peak_heating_1
            and self.peak_heating_2 == other.peak_heating_2
            and self.peak_cooling_1 == other.peak_cooling_1
            and self.peak_cooling_2 == other.peak_cooling_2
        )

    @property
    def max_temperature_air(self) -> float:
        """The maximum air temperature in the dataset."""
        return max(self.temperature_air)

    @property
    def min_temperature_air(self) -> float:
        """The minimum air temperature in the dataset."""
        return min(self.temperature_air)

    @property
    def max_air_temperature_amplitude(self) -> float:
        """The absolute difference between the maximum and minimum air temperature."""
        return self.max_temperature_air - self.min_temperature_air

    @property
    def average_temperature_amplitude(self) -> float:
        """The Average between the maximum and minimum air temperature.
        This value is primarily used to calculate the ground heat loss in PHPP.
        """
        return (self.max_air_temperature_amplitude) / 2

    @property
    def monthly_hours(self) -> List[Tuple[str, int]]:
        """The a ordered (Jan-Dec) list of tuples containing the month-name and the number of hours in that month."""
        return [
            ("JAN", 744),
            ("FEB", 672),
            ("MAR", 744),
            ("APR", 720),
            ("MAY", 744),
            ("JUN", 720),
            ("JUL", 744),
            ("AUG", 744),
            ("SEP", 720),
            ("OCT", 744),
            ("NOV", 720),
            ("DEC", 744),
        ]

    @property
    def monthly_values(self) -> Generator[PhxClimateIterOutput, Any, Any]:
        """A generator that yields the monthly values one at a time for the climate data."""
        for i in range(len(self.monthly_hours)):
            try:
                yield PhxClimateIterOutput(
                    self.monthly_hours[i][0],
                    self.monthly_hours[i][1],
                    self.temperature_air[i],
                    self.temperature_dewpoint[i],
                    self.temperature_sky[i],
                    self.radiation_north[i],
                    self.radiation_east[i],
                    self.radiation_south[i],
                    self.radiation_west[i],
                    self.radiation_global[i],
                )
            except IndexError:
                msg = "Error: Climate data lengths do not match?"
                raise IndexError(msg)


@dataclass
class PhxPHPPCodes:
    country_code: str = "US-United States of America"
    region_code: str = "New York"
    dataset_name = "US0055b-New York"

    def __eq__(self, other) -> bool:
        return (
            self.country_code == other.country_code
            and self.region_code == other.region_code
            and self.dataset_name == other.dataset_name
        )


@dataclass
class PhxSite:
    """The climate and location date for the building's site."""

    display_name: str = "New York"

    source: str = "__unknown__"
    selection: SiteSelection = SiteSelection.USER_DEFINED

    location: PhxLocation = field(default_factory=PhxLocation)
    climate: PhxClimate = field(default_factory=PhxClimate)
    phpp_codes: PhxPHPPCodes = field(default_factory=PhxPHPPCodes)
    ground: PhxGround = field(default_factory=PhxGround)

    energy_factors: PhxSiteEnergyFactors = field(default_factory=PhxSiteEnergyFactors)

    def __eq__(self, other) -> bool:
        return (
            self.display_name == other.display_name
            and self.source == other.source
            and self.selection == other.selection
            and self.location == other.location
            and self.climate == other.climate
            and self.phpp_codes == other.phpp_codes
            and self.ground == other.ground
            and self.energy_factors == other.energy_factors
        )
